#ifndef UTMRF_H_APR_12_2006
#define UTMRF_H_APR_12_2006

#include <cstdlib>
#include <cfloat>
#include <fstream>
#include "mrf.h"
#include "samplesat.h"
#include "random.h"
#include "timer.h"
#include "mln.h"
#include "groundpredicate.h"
#include "maxwalksat.h"
#include "convergencetest.h"
#include "gelmanconvergencetest.h"


//next two used for diagnostic runs
struct ClauseInfoWrtPred{
  int predLocation; //predicate location within clause
  double clauseClassCounts[6];
  
  double getTotals() {
    return clauseClassCounts[0] + clauseClassCounts[1] +
      clauseClassCounts[2] + clauseClassCounts[3];
  }
  
  void printCounts() {
    cout << clauseClassCounts[0] << " " << clauseClassCounts[1];
    cout << " " << clauseClassCounts[2] << " " << clauseClassCounts[3] << endl;
  }

 void printAllCounts() {
    cout << clauseClassCounts[0] << " " << clauseClassCounts[1];
    cout << " " << clauseClassCounts[2] << " " << clauseClassCounts[3];
    cout << " " << clauseClassCounts[4] << " " << clauseClassCounts[5] << endl;
 }
};

struct ClauseWithInfo
{
  Clause *theClause;
  Array<ClauseInfoWrtPred *> theInfo;
};
///////////////////////////////////////////////////////////////////////////


class UTMRF
{
 public:
  static const int AGOOD = 0;
  static const int NAGOOD = 1;
  static const int GENERALIZE = 2;
  static const int SPECIALIZE = 3;
  static const int EXTRANAGOOD = 4;
  static const int EXTRAGENER = 5;
  
  UTMRF(const GroundPredicateHashArray* const& queries, 
	const Array<bool>* const& predicateTruthValues, //used to determine class memberships
	const Domain* const & domain,  const Database * const & db, 
	const MLN* const & mln, const Array<Clause *>& trulyInitialClauses,
	const bool& markHardGndClauses,
	const bool& trackParentClauseWts, const int& memLimit = 0)
  {
    assert (queries->size() == predicateTruthValues->size());

    random_.init(-1);
    cout << "creating diagnostic mrf..." << endl; 
    Timer timer;
    GroundPredicateSet unseenPreds, seenPreds;
    GroundPredicateToIntMap gndPredsMap;
    GroundClauseSet gndClausesSet;

    int allPredsId = (*queries)[0]->getId();

    //This was the original way:
    //const Array<IndexClause*>* ungndClauses = mln->getClausesContainingPred(allPredsId);

    //This is with Jesse's bug fix
    Array<IndexClause*>* ungndClauses = (Array<IndexClause*>*)mln->getClausesContainingPred(allPredsId);
    if (!ungndClauses){
      ungndClauses = new Array<IndexClause*>;
    }

    //we will only look at clauses that are in the truly initial clauses
    Array<IndexClause*>* allowedUngndClauses = new Array<IndexClause*>;
    for (int i = 0; i < ungndClauses->size(); i++) {
      Clause* c = (*ungndClauses)[i]->clause;
   
      bool found = false;
      for (int j = 0; j < trulyInitialClauses.size() && !found; j++) 
		found = trulyInitialClauses[j]->same(c);

      if (found){ 
		allowedUngndClauses->append((*ungndClauses)[i]);
      }
    }

    gndPreds_ = new Array<GroundPredicate*>;
    gndClauses_ = new Array<GroundClause*>;

    ungndClausesWithInfo_ = new Array<ClauseWithInfo*>;
    ungndClausePointers_ = new Array<Array<int>*>;
    predAppearsInGndClauses_ = new Array<Array<int>*>(queries->size());

    blocks_ = new Array<Array<int> >;
    blocks_->growToSize(domain->getNumPredBlocks());
    
    blockEvidence_ = new Array<bool>(*(domain->getBlockEvidenceArray()));
    
    long double memNeeded = 0;
	
      //add GroundPredicates in queries to unseenPreds
    for (int i = 0; i < queries->size(); i++){
      GroundPredicate* gp = (*queries)[i];
      unseenPreds.insert(gp);
      int gndPredIdx = gndPreds_->append(gp);
      assert(i == gndPredIdx);
      assert(gndPredsMap.find(gp) == gndPredsMap.end());
      gndPredsMap[gp] = gndPredIdx;
      assert(gndPredsMap[gp] == gndPredIdx);
      
      Array<int> * forThisPred = new Array<int>;
      predAppearsInGndClauses_->append(forThisPred);
    }

    // If too much memory to build MRF then destroy it
    //cout << "The mem limit is " << memLimit << endl;
    if (memLimit > 0){
      memNeeded = sizeKB();
      if (memNeeded > memLimit){
        for (int i = 0; i < gndClauses_->size(); i++)
          delete (*gndClauses_)[i];
        delete gndClauses_;    
	
        for (int i = 0; i < gndPreds_->size(); i++)
          if ((*gndPreds_)[i]) delete (*gndPreds_)[i];
        delete gndPreds_;
	
        for (int i = 0; i < blocks_->size(); i++)
          (*blocks_)[i].clearAndCompress();
        delete blocks_;
	
        delete blockEvidence_;
        
        throw 1;
      }
    }

    //prepare the ungnd clause with info array 
    //can do it out here because these are all
    //groundings of the same predicate
    for (int i = 0; i < allowedUngndClauses->size(); i++){
      Clause* c = (*allowedUngndClauses)[i]->clause;
      
      ClauseWithInfo* cwi = new ClauseWithInfo;
      cwi->theClause = c;
      ungndClausesWithInfo_->append(cwi);    
    }

    //while there are still unknown preds we have not looked at
    while (!unseenPreds.empty())   
    {
      GroundPredicateSet::iterator predIt = unseenPreds.begin();
      GroundPredicate* pred = *predIt;
      int predIndex = gndPredsMap[pred];

      //for each clause, ground it and find those with unknown truth values,
      //dropping ground preds which do not matter to the clause's truth value
      for (int i = 0; allowedUngndClauses && i < allowedUngndClauses->size(); i++)
      {
        Clause* c = (*allowedUngndClauses)[i]->clause;
		const int clauseId = mln->findClauseIdx(c);  
		assert(clauseId >= 0);	

		//add gnd clauses with unknown truth values to gndClauses_
        const double* parentWtPtr = (trackParentClauseWts) ? c->getWtPtr():NULL;

		//here make unseen preds as they should be

		//the false stands for allPredGndingsAreQueries
        AddGroundClauseStruct agc(&seenPreds, &unseenPreds, gndPreds_,
                                  &gndPredsMap, false, 
                                  &gndClausesSet, gndClauses_,
                                  markHardGndClauses, parentWtPtr,
				  clauseId); //i is the ungndclause Id


		addUnknownDiagnosticGndClauses( pred, c, domain, db, predIndex, 
					        &agc, (*predicateTruthValues)[predIndex], i);
     }
      //clauses with negative wts are flipped in prepareMaxWalksatData()

      //if all the gnd clauses that pred appears in have known truth value,
      //it is not added to gndPreds_ and excluded from the MCMC network

      unseenPreds.erase(predIt);
      seenPreds.insert(pred);
    }//while (!unseenPreds.empty())

    cout << "number of grounded predicates = " << gndPreds_->size() << endl;
    cout << "number of grounded clauses = " << gndClauses_->size() << endl;
    if (gndClauses_->size() == 0)
      cout<< "Markov blankets of query ground predicates are empty" << endl;    
     
    gndPreds_->compress();
    gndClauses_->compress();

    cout <<"Time taken to construct MRF = ";
    Timer::printTime(cout,timer.time());
    cout << endl;

  }


  //a shorter version ot be used for the do-we-have-the-right-clauses diagnosis
  //doesn't instantiate all the datastructures
  UTMRF(const GroundPredicateHashArray* const& queries, 
	const Domain* const & domain,  const Database * const & db, 
	const MLN* const & mln)
  {
    //int memLimit = 0;

    random_.init(-1);
    cout << "creating diagnostic mrf... (short version)" << endl; 
    Timer timer;
    GroundPredicateSet unseenPreds, seenPreds;
    GroundPredicateToIntMap gndPredsMap;
    GroundClauseSet gndClausesSet;

   
    //  cout << "The size of the queries is " << queries->size() << endl; cout.flush();

    int allPredsId = (*queries)[0]->getId();

    //This is the original way
    //const Array<IndexClause*>* ungndClauses = mln->getClausesContainingPred(allPredsId);

    //This is with Jesse's bug fix:
    Array<IndexClause*>* ungndClauses = (Array<IndexClause*>*)mln->getClausesContainingPred(allPredsId);
    if (!ungndClauses){
      ungndClauses = new Array<IndexClause*>;
    }

    gndPreds_ = new Array<GroundPredicate*>;
    gndClauses_ = new Array<GroundClause*>;

    ungndClausesWithInfo_ = NULL; //new Array<ClauseWithInfo*>;
    ungndClausePointers_ = NULL; //new Array<Array<int>*>;
    predAppearsInGndClauses_ = NULL; //new Array<Array<int>*>(queries->size());

    blocks_ = new Array<Array<int> >;
    blocks_->growToSize(domain->getNumPredBlocks());
    blockEvidence_ = new Array<bool>(*(domain->getBlockEvidenceArray()));
    
      //add GroundPredicates in queries to unseenPreds
    for (int i = 0; i < queries->size(); i++){
      GroundPredicate* gp = (*queries)[i];
      unseenPreds.insert(gp);
      int gndPredIdx = gndPreds_->append(gp);
      assert(gndPredsMap.find(gp) == gndPredsMap.end());
      gndPredsMap[gp] = gndPredIdx;
    }

    //while there are still unknown preds we have not looked at
    while (!unseenPreds.empty())   
    {
      GroundPredicateSet::iterator predIt = unseenPreds.begin();
      GroundPredicate* pred = *predIt;
   
      //for each clause, ground it and find those with unknown truth values,
      //dropping ground preds which do not matter to the clause's truth value
      for (int i = 0; ungndClauses && i < ungndClauses->size(); i++)
      {
        Clause* c = (*ungndClauses)[i]->clause;

	if (c->getWt() == 0) continue;

	const int clauseId = mln->findClauseIdx(c);  
	assert(clauseId >= 0);	

	//add gnd clauses with unknown truth values to gndClauses_
        const double* parentWtPtr = NULL; //(trackParentClauseWts) ? c->getWtPtr():NULL;
	
	//here make unseen preds as they should be

	//the false stands for allPredGndingsAreQueries
        AddGroundClauseStruct agc(&seenPreds, &unseenPreds, gndPreds_,
                                  &gndPredsMap, false, 
                                  &gndClausesSet, gndClauses_,
                                  false, parentWtPtr,
				  clauseId); //i is the ungndclause Id

	UTMRF::addUnknownGndClauses(pred, c, domain, db, false, &agc);

      }
      //clauses with negative wts are flipped in prepareMaxWalksatData()

      unseenPreds.erase(predIt);
      seenPreds.insert(pred);
    }//while (!unseenPreds.empty())

    if (gndClauses_->size() == 0)
      cout<< "Markov blankets of query ground predicates are empty" << endl;    
 
    gndPreds_->compress();
    gndClauses_->compress();

    cout <<"Time taken to construct MRF = ";
    Timer::printTime(cout,timer.time());
    cout << endl;
  }



 //L:Made it not static
  //Do not delete the clause and truncClause argument.
    //This function is tightly bound to Clause::createAndAddUnknownClause().
  void addUnknownGndClause(const AddGroundClauseStruct* const & agcs, 
			   const Clause* const & clause,
			   const Clause* const & truncClause,
			   const bool& isHardClause, int ungndClauseId)
  {
    const GroundPredicateSet* seenPreds     = agcs->seenPreds;
    GroundPredicateSet*       unseenPreds   = agcs->unseenPreds;
    Array<GroundPredicate*>*  gndPreds      = agcs->gndPreds;
    GroundPredicateToIntMap*  gndPredsMap   = agcs->gndPredsMap; 
    //const Array<int>* allGndingsAreQueries  = agcs->allPredGndingsAreQueries;
    GroundClauseSet*          gndClausesSet = agcs->gndClausesSet;
    Array<GroundClause*>*     gndClauses    = agcs->gndClauses;
    const bool markHardGndClauses           = agcs->markHardGndClauses;
    const double* parentWtPtr               = agcs->parentWtPtr;
	

    GroundClause *gndClause = new GroundClause(truncClause);
    if (markHardGndClauses && isHardClause) gndClause->setWtToHardWt();

    //we are ok with clauses with 0 weight for the purposes of diagnosis. 
    if (gndClause->getWt() == 0)
      gndClause->setWt(0.1);

    //assert(gndClause->getWt() != 0);


    GroundClauseSet::iterator iter = gndClausesSet->find(gndClause);
      //if the unknown clause is not in gndClauses
    if (iter == gndClausesSet->end())
    {
      gndClausesSet->insert(gndClause);
      gndClauses->append(gndClause);
      
      if (ungndClauseId >= 0){
	recordUngndClauseForGndClause(gndClause, ungndClauseId, gndClauses->size() - 1);
      }

      //gndClause's wt is set when it was constructed
      if (parentWtPtr) 
      { 
        gndClause->appendParentWtPtr(parentWtPtr);
        assert(gndClause->getWt() == *parentWtPtr);
      }

        // add the unknown predicates of the clause to unseenPreds if 
        // the predicates are already not in it
      for (int j = 0; j < gndClause->getNumGroundPredicates(); j++)
      {
        GroundPredicate* gp =(GroundPredicate*)gndClause->getGroundPredicate(j);
        assert(seenPreds->find(gp) == seenPreds->end());
          // if the ground predicate is not in unseenPreds
        GroundPredicateSet::iterator it = unseenPreds->find(gp);
        if (it == unseenPreds->end())
        {
          //cout << "\tinserting into unseen pred: ";  
          //pred->print(cout, domain); cout << endl;
          unseenPreds->insert(gp);
          //commented out: done when gndClause is initialized
          //gp->appendGndClause(gndClause,predSense);              
          int gndPredIdx = gndPreds->append(gp);
          assert(gndPredsMap->find(gp) == gndPredsMap->end());
          (*gndPredsMap)[gp] = gndPredIdx;
          gndClause->setGroundPredicateIndex(j, gndPredIdx);
        }
        else
        {
          GroundPredicate* origGndPred = *it;
          gndClause->replaceGroundPredicate(j, origGndPred);
          bool predSense = gndClause->getGroundPredicateSense(j);
          bool ok = origGndPred->appendGndClause(gndClause, predSense);
          assert(ok); ok = true; // avoid compilation warning
          
            // origGndPred must already be in gndPreds
          int gndPredIdx = (*(gndPredsMap->find(origGndPred))).second;
          assert(gndPredIdx >= 0);
          gndClause->setGroundPredicateIndex(j, gndPredIdx);
        }
      }
    }
    else
    { 
      // gndClause has appeared before, so just accumulate its weight
      if (ungndClauseId >= 0){
	recordUngndClauseForGndClause(gndClause, ungndClauseId);
      }
      (*iter)->addWt(gndClause->getWt());
      if (parentWtPtr) (*iter)->appendParentWtPtr(parentWtPtr);
      gndClause->deleteAllGndPreds();
      delete gndClause;
    }
  } //addUnknownGndClause()


  /**
   * Computes and returns size of the mrf in kilobytes
   */
  long double sizeKB()
  {
      // # of ground clauses times memory for a ground clause +
      // # of ground predicates times memory for a ground predicate
    long double size = 0;
    for (int i = 0; i < gndClauses_->size(); i++)
      size += (*gndClauses_)[i]->sizeKB();
    for (int i = 0; i < gndPreds_->size(); i++)
      size += (*gndPreds_)[i]->sizeKB();

    return size;    
  }

    //Do not delete the clause and truncClause argument.
    //This function is tightly bound to Clause::createAndAddUnknownClause().
  static void addUnknownGndClause(const AddGroundClauseStruct* const & agcs, 
                                  const Clause* const & clause,
                                  const Clause* const & truncClause,
                                  const bool& isHardClause)
  {
    const GroundPredicateSet* seenPreds     = agcs->seenPreds;
    GroundPredicateSet*       unseenPreds   = agcs->unseenPreds;
    Array<GroundPredicate*>*  gndPreds      = agcs->gndPreds;
    GroundPredicateToIntMap*  gndPredsMap   = agcs->gndPredsMap; 
    const Array<int>* allGndingsAreQueries  = agcs->allPredGndingsAreQueries;
    GroundClauseSet*          gndClausesSet = agcs->gndClausesSet;
    Array<GroundClause*>*     gndClauses    = agcs->gndClauses;
    const bool markHardGndClauses           = agcs->markHardGndClauses;
    const double* parentWtPtr               = agcs->parentWtPtr;
    const int clauseId                      = agcs->clauseId;
	

    //check none of the grounded clause's predicates have been seen before
    //if any of them have been seen before, this clause has been created 
    //before (for that seen predicate), and can be ignored

      //check the untruncated ground clause whether any of its predicates
      //have been seen before
    GroundClause* gndClause = new GroundClause(clause);
    bool seenBefore = false;
    for (int j = 0; j < gndClause->getNumGroundPredicates(); j++)
    {
      GroundPredicate* gp = (GroundPredicate*) gndClause->getGroundPredicate(j);
      if (seenPreds->find(gp) !=  seenPreds->end() ||
          (allGndingsAreQueries && (*allGndingsAreQueries)[gp->getId()] > 1) )
      { 
        seenBefore = true;
        break;
      }
    }

    gndClause->deleteAllGndPreds();
    delete gndClause;
    if (seenBefore) return;

    gndClause = new GroundClause(truncClause);
    if (markHardGndClauses && isHardClause) gndClause->setWtToHardWt();
    //assert(gndClause->getWt() != 0);


    GroundClauseSet::iterator iter = gndClausesSet->find(gndClause);
      //if the unknown clause is not in gndClauses
    if (iter == gndClausesSet->end())
    {
	  gndClausesSet->insert(gndClause);
      gndClauses->append(gndClause);
      //gndClause's wt is set when it was constructed
      if (parentWtPtr) 
      { 
        gndClause->appendParentWtPtr(parentWtPtr);
		gndClause->incrementClauseFrequency(clauseId,1);		
		assert(gndClause->getWt() == *parentWtPtr);
      }

        // add the unknown predicates of the clause to unseenPreds if 
        // the predicates are already not in it
      for (int j = 0; j < gndClause->getNumGroundPredicates(); j++)
      {
        GroundPredicate* gp =(GroundPredicate*)gndClause->getGroundPredicate(j);
        assert(seenPreds->find(gp) == seenPreds->end());
          // if the ground predicate is not in unseenPreds
        GroundPredicateSet::iterator it = unseenPreds->find(gp);
        if (it == unseenPreds->end())
        {
          
          unseenPreds->insert(gp);
          //commented out: done when gndClause is initialized
          //gp->appendGndClause(gndClause,predSense);              
          int gndPredIdx = gndPreds->append(gp);
          assert(gndPredsMap->find(gp) == gndPredsMap->end());
          (*gndPredsMap)[gp] = gndPredIdx;
          gndClause->setGroundPredicateIndex(j, gndPredIdx);
        }
        else
        {
          GroundPredicate* origGndPred = *it;
          gndClause->replaceGroundPredicate(j, origGndPred);
          bool predSense = gndClause->getGroundPredicateSense(j);
          bool ok = origGndPred->appendGndClause(gndClause, predSense);
          assert(ok); ok = true; // avoid compilation warning
          
            // origGndPred must already be in gndPreds
          int gndPredIdx = (*(gndPredsMap->find(origGndPred))).second;
          assert(gndPredIdx >= 0);
          gndClause->setGroundPredicateIndex(j, gndPredIdx);
        }
      }
    }
    else
    {  // gndClause has appeared before, so just accumulate its weight
      (*iter)->addWt(gndClause->getWt());

	  if (parentWtPtr)
	  {
		(*iter)->appendParentWtPtr(parentWtPtr);
        (*iter)->incrementClauseFrequency(clauseId,1); 
	  }

	  gndClause->deleteAllGndPreds();
      delete gndClause;
    }
  } //addUnknownGndClause()



  ~UTMRF()
  {
  	/*
    for (int i = 0; i < gndClauses_->size(); i++)
      delete (*gndClauses_)[i];
    delete gndClauses_;    

    // for (int i = 0; i < gndPreds_->size(); i++)
    //  if ((*gndPreds_)[i]) delete (*gndPreds_)[i];
    delete gndPreds_;
    
    for (int i = 0; i < blocks_->size(); i++)
      (*blocks_)[i].clearAndCompress();
    delete blocks_;
    
    delete blockEvidence_;

    if (ungndClausesWithInfo_ != NULL) {
      for (int i = 0; i < ungndClausesWithInfo_->size(); i++){
	//the Clause with info contains the clause and the info. We are
	//not deleting the clause because we don't own it.
	for (int j = 0; j < (*ungndClausesWithInfo_)[i]->theInfo.size(); j++){
	  delete (*ungndClausesWithInfo_)[i]->theInfo[j];
	}
      
	delete (*ungndClausesWithInfo_)[i];
      }
    
      for (int i = 0; i < predAppearsInGndClauses_->size(); i++)
	delete (*predAppearsInGndClauses_)[i];  
      delete predAppearsInGndClauses_;  
    
      for (int i = 0; i < ungndClausePointers_->size(); i++) 
	delete (*ungndClausePointers_)[i];
      delete ungndClausePointers_;
    }
    */
  }

 //L: if we know the id of the gndClause in gndClauses_, we pass it as the last argument
  void recordUngndClauseForGndClause(GroundClause * gndClause, int ungndClauseId, 
				     int gndClausesId = -1) {
    //make sure we have entries for pointers for new gnd clauses
    for (int i = ungndClausePointers_->size(); i < gndClauses_->size(); i++) {
      Array<int> *newArray = new Array<int>;
      ungndClausePointers_->append(newArray);
    }
    
    //find the clause in the array of gnd clauses so we get the id in the array
    int i = 0;
    if (gndClausesId == -1) {
      for (i = 0; i < gndClauses_->size(); i++){
	if ((*gndClauses_)[i]->same(gndClause))
	  break;
      }
    }
    else
      i = gndClausesId;

    assert (i < gndClauses_->size());
    //add the id of the ungnd clause if it's not already there
    Array<int>* thisClauseIdxs = (*ungndClausePointers_)[i];

    bool found = false;
    for (int j = 0; j < thisClauseIdxs->size() && !found; j++) 
      found = ((*thisClauseIdxs)[j] == ungndClauseId);
      
    if (!found)
      thisClauseIdxs->append(ungndClauseId);
  }

  void deleteGndClausesIntArrReps() 
  {
    for (int i = 0; i < gndClauses_->size(); i++)
      (*gndClauses_)[i]->deleteIntArrRep();    
  }


  void deleteGndPredsGndClauseSets()
  {
    for (int i = 0; i < gndPreds_->size(); i++)
      (*gndPreds_)[i]->deleteGndClauseSet();
  }  

  
  void setGndClausesWtsToSumOfParentWts()
  {
    for (int i = 0; i < gndClauses_->size(); i++)
      (*gndClauses_)[i]->setWtToSumOfParentWts();
  }


  const Array<GroundPredicate*>* getGndPreds() const { return gndPreds_; }

  const Array<GroundClause*>* getGndClauses() const { return gndClauses_; }

  void initGndPredsTruthValues(const int& numChains)
  {
    for (int i = 0; i < gndPreds_->size(); i++)
      (*gndPreds_)[i]->initTruthValues(numChains);
  }

  //needed for the "small" diagnosis
  void calculateWeightsOfGroundings() {   
   //initialize gndPreds' truthValues & wts, and gndClauses' numSatLiterals
    for (int i = 0; i < gndPreds_->size(); i++) {
      (*gndPreds_)[i]->initTruthValuesAndWts(1);
    }

    initNumSatLiterals(1);

    GroundPredicateHashArray affectedGndPreds;

    // The following 5 lines were not in the original code. Need to
    // check that something funky doesn't go on.
    for (int i = 0; i < gndPreds_->size(); i++) 
      affectedGndPreds.append((*gndPreds_)[i], gndPreds_->size());
    
    updateWtsForAffectedGndPreds(affectedGndPreds, 0);        
  }


  // ========================================================================= //
    // If saveSapce is set to true, the gndPredSenses_ and gndPredIndexes_
    // fields of the GroundClauses in gndClauses_ will be deleted in the 
    // function. This means that the function can only be correctly called once.
    // Returns the number of samples drawn per predicate.
  // ========================================================================= //
  int diagnosticGibbsSampling(const GibbsParams& gibbsParams, 
			      const MaxWalksatParams* const & maxWalksatParams,
			      const bool& saveSpace, const Domain* const & domain, 
			      const Array<bool>* const & predicateTruthValues)
  {
    // this occurs when all query predicates don't appear in any UNKNOWN clause 
    if (gndClauses_->size()==0){ 
      //but first fix the counts
      for (int cl = 0; cl < ungndClausesWithInfo_->size(); cl++) {
		ClauseWithInfo *currClause = (*ungndClausesWithInfo_)[cl];
		for (int i = 0; i < currClause->theInfo.size(); i++){
		  ClauseInfoWrtPred* currcip = currClause->theInfo[i];
		  currcip->clauseClassCounts[NAGOOD] = 
		    currcip->clauseClassCounts[EXTRANAGOOD];
		  currcip->clauseClassCounts[GENERALIZE] = 
		    currcip->clauseClassCounts[EXTRAGENER];
		}
      }    
      return 0;
    }

    Timer timer;
    double begSec;
      //extract Gibbs sampling parameters
    int numChains = gibbsParams.numChains;
    WalksatType walksatType = gibbsParams.walksatType;
    double gamma = gibbsParams.gamma;
    double epsilonFrac = gibbsParams.epsilonFrac;
    int samplesPerConvergenceTest = gibbsParams.samplesPerConvergenceTest;
    int burnMinSteps = gibbsParams.burnMinSteps;
    int burnMaxSteps = gibbsParams.burnMaxSteps;
    int gibbsMinSteps = gibbsParams.gibbsMinSteps;
    int gibbsMaxSteps = gibbsParams.gibbsMaxSteps;
    double fracConverged = gibbsParams.fracConverged;
    int maxSeconds = gibbsParams.maxSeconds;
    
    int numGndPreds = gndPreds_->size();

      //initialize gndPreds' truthValues & wts, and gndClauses' numSatLiterals
    for (int i = 0; i < gndPreds_->size(); i++)
      (*gndPreds_)[i]->initTruthValuesAndWts(numChains);

    begSec = timer.time();
    cout << "initializing ground predicates values "; 
    if (walksatType == MAXWALKSAT)
    {
      assert(maxWalksatParams);
      cout << "with MaxWalksat" << endl;
      maxWalksatInitGndPredsTruthValues(numChains, maxWalksatParams);
    }
    else
    {
      cout << "randomly " << endl;
      randomInitGndPredsTruthValues(numChains);
    }
    cout << "initialization took "; Timer::printTime(cout, timer.time()-begSec);
    cout << endl;

    for (int i = 0; i < gndClauses_->size(); i++)
      (*gndClauses_)[i]->initNumSatLiterals(numChains);

    //initialize convergence test
    GelmanConvergenceTest** burnConvergenceTests;
    ConvergenceTest** gibbsConvergenceTests;
    initConvergenceTests(burnConvergenceTests, gibbsConvergenceTests, 
                         gamma, epsilonFrac, numGndPreds, numChains);

    int numSamplesPerPred = 0, isamp = -1;
    bool done      = false;
    bool burningIn = true;
  
    double secondsElapsed = 0;
    double startTimeSec = timer.time();
    double currentTimeSec;

    GroundPredicateHashArray affectedGndPreds;

    // The following 5 lines were not in the original code. Need to
    // check that something funky doesn't go on.
    for (int i = 0; i < gndPreds_->size(); i++) 
      affectedGndPreds.append((*gndPreds_)[i], gndPreds_->size());
    for (int c = 0; c < numChains; c++) 
      updateWtsForAffectedGndPreds(affectedGndPreds, c);        
    affectedGndPreds.clear();


    cout << "running Gibbs sampling..." << endl;
      // start burn in and Gibbs sampling
    while(!done) 
    {
      ++isamp;

      if (isamp % samplesPerConvergenceTest == 0)
      { 
        currentTimeSec = timer.time();
        secondsElapsed = currentTimeSec-startTimeSec;
        cout << "Sample (per pred per chain) " << isamp << ", time elapsed = ";
        Timer::printTime(cout,secondsElapsed); cout << endl;
      }

      // For each chain, for each node, generate the node's new truth value
      for (int c = 0; c < numChains; c++) 
      {
		for (int i = 0; i < gndPreds_->size(); i++) 
	    {
		  bool newAssignment 
	            = genTruthValueForProb((*gndPreds_)[i]->getProbability(c));
	
            // if gndPred is flipped, do updates & find all affected gndPreds
          if (newAssignment != (*gndPreds_)[i]->getTruthValue(c)) 
          {            
            (*gndPreds_)[i]->setTruthValue(c,newAssignment);
            affectedGndPreds.clear();
            gndPredFlippedUpdates(i, c, affectedGndPreds);
            updateWtsForAffectedGndPreds(affectedGndPreds, c);
          }

            // if in actual gibbs sampling phase, track the num of times
            // the ground predicate is set to true
          if (!burningIn && newAssignment) (*gndPreds_)[i]->incrementNumTrue();

		  //L: this is where the difference between diagnostic and
		  //non-diagnostic appears. For each formula, in which the
		  //current gnd Predicate participates, we need to reevaluate
		  //the formula and see in what category it falls.
		  if (!burningIn) {
		    bool currPredValue = (*gndPreds_)[i]->getTruthValue(c);
		    Array<int> * indicesOfClauses = (*predAppearsInGndClauses_)[i];
		  
		    for (int cli = 0; cli < indicesOfClauses->size(); cli++) {
		      int currClauseIndex = (*indicesOfClauses)[cli];
		      GroundClause * currentGC = (*gndClauses_)[currClauseIndex];
		      
		      //cout << "The current ground clause is "; currentGC->print(cout, domain); cout << endl; 
	
		      //go over each ungnd clause which can be grounded like this  
		      Array <int>* participatingClauses = (*ungndClausePointers_)[currClauseIndex];
		      
		      for (int k = 0; k < participatingClauses->size(); k++){
				int clauseId = (*participatingClauses)[k];
				Clause * currUngndClause = (*ungndClausesWithInfo_)[clauseId]->theClause;
			
				//cout << "The current ungnd clause is "; currUngndClause->print(cout, domain); cout << endl;
				for (int t = 0; t < currUngndClause->getNumPredicates(); t++){
				  if (currUngndClause->getPredicate(t)->canBeGroundedAs((*gndPreds_)[i])){
				    
				    //t is my position in the ungnd clause
				    
				    bool currPredSenseInCl = (*ungndClausesWithInfo_)[clauseId]->
				      theClause->getPredicate(t)->getSense();
				    double weight = (*ungndClausesWithInfo_)[clauseId]->theClause->getWt();
				    int theClass = -1;
				    
				    // now we need to evaluate this clause
				    if (currentGC->getNumSatLiterals(c) > 0 &&
					currPredValue != currPredSenseInCl) {//ie it's trivially sat.
				      if ( (*predicateTruthValues)[i] != currPredSenseInCl ){
					if (weight >= 0) theClass = NAGOOD;
					else theClass = GENERALIZE;
				      }
				      else{
					if (weight >= 0) theClass = GENERALIZE;
					else theClass = NAGOOD;
				      }
				    }
				    else{ //the clause is not trivially sat
				      if ((*predicateTruthValues)[i] == currPredSenseInCl){
					if (weight >= 0) theClass = AGOOD;
					else theClass = SPECIALIZE;
				      }
				      else{
					if (weight >= 0) theClass = SPECIALIZE;
					else theClass = AGOOD;
				      }
				    }
				    assert(theClass >= 0 && theClass < 4);
		
				    //cout << "The class is " << theClass << endl;
				    recordClass(theClass, clauseId, t, 1);
		
	   
				  }  
				}
			   }
		    }
		  }
        }
		if (!burningIn) numSamplesPerPred++;
      }//end of chain

  
        // add current truth values to the convergence testers
      for (int i = 0; i < gndPreds_->size(); i++) 
      {        
        const bool* vals = (*gndPreds_)[i]->getTruthValues()->getItems();
          //WARNING: implicit cast from bool* to double*
        if (burningIn) burnConvergenceTests[i]->appendNewValues(vals);
        else           gibbsConvergenceTests[i]->appendNewValues(vals);
      }


      if (isamp % samplesPerConvergenceTest != 0) continue;      
      if (burningIn) 
      {
          // use convergence criteria stated in "Probability and Statistics",
          // DeGroot and Schervish
        bool burnConverged 
          = GelmanConvergenceTest::checkConvergenceOfAll(burnConvergenceTests,
                                                         gndPreds_->size(),
                                                         true);
        if (   (isamp >= burnMinSteps && burnConverged)
            || (burnMaxSteps >= 0 && isamp >= burnMaxSteps) 
            || (maxSeconds > 0 && secondsElapsed >= maxSeconds))
        {
          cout << "Done burning. " << isamp << " samples per pred per chain (" 
               << (burnConverged? "converged":"didn't converge") 
               <<" at total of " << numChains*isamp << " samples per pred)" 
               << endl;
          burningIn = false;
          isamp=0;          
        }
      }
      else 
      {  // going actual gibbs sampling
        bool gibbsConverged 
          = ConvergenceTest::checkConvergenceOfAtLeast(gibbsConvergenceTests, 
                                                       gndPreds_->size(), isamp,
                                                       fracConverged, true);
        if (   (isamp >= gibbsMinSteps && gibbsConverged) 
            || (gibbsMaxSteps >= 0 && isamp >= gibbsMaxSteps) 
            || (maxSeconds > 0 && secondsElapsed >= maxSeconds)) 
        {
          cout << "Done Diagnostic Gibbs sampling. " << isamp 
               << " samples per pred per chain ("
               << (gibbsConverged? "converged":"didn't converge") 
               <<" at total of " << numSamplesPerPred << " samples per pred)" 
               << endl;
          done = true;
        }
      }
      cout.flush();
    } // while (!done);    


    cout<< "Time taken for Diagnostic Gibbs sampling = "; 
    Timer::printTime(cout, timer.time()-startTimeSec); cout << endl;

      // update gndPreds_ probability that it is true
    for (int i = 0; i < gndPreds_->size(); i++)
      (*gndPreds_)[i]->setProbTrue((*gndPreds_)[i]->getNumTrue()
                                   / numSamplesPerPred);
    //L:Here take care of the extraones They are weighted by the
    //percentage of time they were of each type. We can do it this way
    //because we know for sure that if they didn't apply, then the
    //formula was not in the mrf for that particular ground
    //predicate. So from that predicate's point of view, they could
    //only have contributed towards one of the "doesn't apply
    //categories"

    for (int cl = 0; cl < ungndClausesWithInfo_->size(); cl++) {
      ClauseWithInfo *currClause = (*ungndClausesWithInfo_)[cl];
      for (int i = 0; i < currClause->theInfo.size(); i++){
		ClauseInfoWrtPred* currcip = currClause->theInfo[i];

		currcip->clauseClassCounts[NAGOOD] += 
		  numSamplesPerPred * currcip->clauseClassCounts[EXTRANAGOOD];
		currcip->clauseClassCounts[GENERALIZE] += 
		  numSamplesPerPred * currcip->clauseClassCounts[GENERALIZE];

		cout << "after adding in the extras we have \n";
		for (int i = 0; i < ungndClausesWithInfo_->size(); i++){
		  ClauseWithInfo * currClauseWI = (*ungndClausesWithInfo_)[i];
		  cout << "__________________________________\n";
		  currClauseWI->theClause->print(cout, domain); cout << endl;
		  for (int j = 0; j < currClauseWI->theInfo.size(); j++){
		    currClauseWI->theInfo[j]->printAllCounts();
		  }
		}
		
		double newTotals = currcip->getTotals();
		for(int j = 0; j < 4; j++) 
		  currcip->clauseClassCounts[j] = 
		    currcip->clauseClassCounts[j]/newTotals;
	
		cout << "printing the clauses with counts\n";
		for (int i = 0; i < ungndClausesWithInfo_->size(); i++){
		  ClauseWithInfo * currClauseWI = (*ungndClausesWithInfo_)[i];
		  cout << "__________________________________\n";
		  currClauseWI->theClause->print(cout, domain); cout << endl;
		  for (int j = 0; j < currClauseWI->theInfo.size(); j++){
		    currClauseWI->theInfo[j]->printAllCounts();
		  }
		}
      }
    }

    deleteConvergenceTests(burnConvergenceTests, gibbsConvergenceTests, 
                           numGndPreds);    
    return numSamplesPerPred;
  }

  //L: Similar to addUnknownGndClauses but specially tailored to the diagnostic task
   void addUnknownDiagnosticGndClauses(const GroundPredicate* const& queryGndPred,
				       Clause* const & c, const Domain* const & domain, 
				       const Database* const & db, int predId,
				       const AddGroundClauseStruct* const & agcs,
				       bool boolTV, int ungndClauseId)
   {
     int clauseId = ungndClauseId;
     double clauseWt = c->getWt();
     for (int i = 0; i < c->getNumPredicates(); i++){
       if (c->getPredicate(i)->canBeGroundedAs(queryGndPred)){
	 
	 int initialGndClSize = gndClauses_->size();

	 int numTrivial = c->addUnknownClauses_ut(domain, db, i, queryGndPred,
						  agcs, this, ungndClauseId);
	 
	 //record which gnd clauses this predicate appears in so that
	 //we can keep the information about its position within the
	 //unground clause
	 for (int j = initialGndClSize; j < gndClauses_->size(); j++)
	   (*predAppearsInGndClauses_)[predId]->append(j);

	 if( numTrivial > 0 ){
	   bool predSense = c->getPredicate(i)->getSense();
	   
	   if (clauseWt >= 0){
	     //If the sense agrees with the conclusion, the rule could
	     //have been correct if it applied. So it needs to be
	     //generalized
	     if (boolTV == predSense){
	       recordClass(EXTRAGENER, clauseId, i, numTrivial);
	     }
	     //Otherwise, it was correct because if it had applied, it
	     //would have given the wrong conclusion.
	     else {
	       recordClass(EXTRANAGOOD, clauseId, i, numTrivial);
	     }
	   }
	   else{ //if negative wt, trend reverses.
	     if (boolTV == predSense) {
	       recordClass(EXTRANAGOOD, clauseId, i, numTrivial);
	     }
	     else {
	       recordClass(EXTRAGENER, clauseId, i, numTrivial);
	     }
	   }
	 }
       }
     }
   }

   void recordClass(int theClass, int clauseId, int predIdInClause, int incrementBy)
   {
     ClauseWithInfo* theClauseWithInfo = (*ungndClausesWithInfo_)[clauseId];
     //first check if we have a record running for this clause/pred combination
     bool found = false;
     int i;
     for (i = 0; !found && i < theClauseWithInfo->theInfo.size(); i++)
       found = (theClauseWithInfo->theInfo[i]->predLocation == predIdInClause);
     
     if (found) 
       theClauseWithInfo->theInfo[i-1]->clauseClassCounts[theClass] += incrementBy;
     else {
       ClauseInfoWrtPred *cwp = new ClauseInfoWrtPred;
       cwp->predLocation = predIdInClause;
       for (int i = 0; i < 6; i++) cwp->clauseClassCounts[i] = 0;
       cwp->clauseClassCounts[theClass] += incrementBy;
       theClauseWithInfo->theInfo.append(cwp);
     } 
   }

   void recordClassForGndClause(int theClass, GroundClause * gndClause, 
				int predIdInClause, int incrementBy, int gndClauseIdx){
     Array <int>* ungndClauseIdxs = (*ungndClausePointers_)[gndClauseIdx];
     assert(ungndClauseIdxs != NULL);

     for (int i = 0; i < ungndClauseIdxs->size(); i++)
       recordClass(theClass, (*ungndClauseIdxs)[i], predIdInClause, incrementBy);
   }

void printDiagnosticDataStructures(const Domain * const & domain) 
  {
    cout << "-----Ground Predicates----\n";
    for (int i = 0; i < gndPreds_->size(); i++) {
      (*gndPreds_)[i]->print(cout, domain); cout << endl; cout.flush();
    }
    cout << "----Finished with Ground Predicates\n\n"; cout.flush();

    cout << "-----Ground clauses and UngndClause Indices\n";
    for (int i = 0; i < gndClauses_->size(); i++) {
      (*gndClauses_)[i]->print(cout, domain); cout << endl; cout.flush();

      cout << "i is " << i << " and the size of ungndClausePointers is " << ungndClausePointers_->size();
      cout << endl; cout.flush();

      cout << "This is a grounding of the following ungnd clauses:\n";
      Array<int>* ungndPointers = (*ungndClausePointers_)[i];
      for (int j = 0; j < ungndPointers->size(); j++)
	cout << (*ungndPointers)[j] << " " << endl; cout.flush();
      cout << endl;
    }
    cout << "----Finished with Ground clauses\n\n"; cout.flush();

    cout << "-----Unground clauses with info\n"; cout.flush();
    for (int i = 0; i < ungndClausesWithInfo_->size(); i++) {
      ClauseWithInfo * cwi = (*ungndClausesWithInfo_)[i];
      cout << "Clause " << i << " is: "; 
      cwi->theClause->print(cout, domain); cout << endl; cout.flush();

      cout << "The info is: " << endl;

      for (int j = 0; j < cwi->theInfo.size(); j++) { cout.flush();
	ClauseInfoWrtPred *cip = cwi->theInfo[j];
	cout << "Pred location is" << cip->predLocation;
	cout << "Printing class counts...\n";
	for (int k = 0; k < 6; k++)
	  cout << k << ": " << cip->clauseClassCounts[k] << endl;
      }
    }
    cout << "----Finished with ungnd clauses with info \n\n";

    cout << "-----PredAppearsInGndClause" << endl;
    for (int i = 0; i < predAppearsInGndClauses_->size(); i++){
      Array<int>* arrayForCurrPred = (*predAppearsInGndClauses_)[i];
      cout << "The predicate is "; (*gndPreds_)[i]->print(cout, domain);
      cout << endl;
      
      for (int j = 0; j < arrayForCurrPred->size(); j++){
	cout << "The clause is "; (*gndClauses_)[(*arrayForCurrPred)[j]]->print(cout, domain);
	cout << endl; cout.flush();
      }
    }
    cout << "----Done checking predAppearsInGndClause" << endl;cout.flush();
  }
 
  //get the clauses with info; must be in diagnostic mode returns a
  //pointer to the data structure in this class. do not delete either
  //the data structure or the mrf object before we are totally done
 public:

  Array<ClauseWithInfo*>* getUngndClausesWithInfo(){
    return ungndClausesWithInfo_;
  }
  
  //returns true if MaxWalksat succeeded; else returns false
  bool runMaxWalksat(const int& chainIdx, const bool& randomInitIfFail,
                     const MaxWalksatParams* const & params, 
                     double& hardClauseWt)
  {
    cout << "running MaxWalksat..." << endl;

    MaxWalksatData* data = prepareMaxWalksatData();
    hardClauseWt = data->hardClauseWt;
    MaxWalksat mws(blocks_, blockEvidence_);
    Array<Array<bool>*> solutions;
    mws.sample(data, chainIdx, solutions, params, blocks_, blockEvidence_);
    
    cout << "MaxWalksat sampling for chain " << chainIdx << "..." << endl;
    
    bool good = true;
    if (solutions.size() == 0) 
    {
      cout << "MaxWalksat was unable to find a solution." << endl;
      good = false;
    }
    else
    if (solutions.size() > 1) 
    {
      cout << "MaxWalksat unexpectedly returned multiple solutions." << endl;
      good = false;
    }

    if (gndPreds_->size() != data->numGndPreds) 
    {
      cout << "Error in MRF: gndPreds_->size() != data->numGndPreds." << endl;
      good = false;
    }

    if (solutions.size() == 1 &&
        gndPreds_->size() != solutions[0]->size()) 
    {
      cout << "Error in MRF: gndPreds_->size() != solutions[0]->size()." <<endl;
      good = false;
    }

 
    if (good)
    {
      Array<bool>* solution = solutions[0];
      for (int i = 0; i < solution->size(); i++)
        (*gndPreds_)[i]->setTruthValue(chainIdx, (*solution)[i] );
    }
    else 
    if (randomInitIfFail)
    {
      cout << "Randomly initializing ground predicates' truth values."
           << endl;
      for (int i = 0; i < gndPreds_->size(); i++) 
        (*gndPreds_)[i]->setTruthValue(chainIdx, genTruthValueForProb(0.5) );
    }
    

    for (int i = 0; i < solutions.size(); i++)
      delete solutions[i];

    for (int i = 0; i < data->clauses.size(); i++)
      delete (data->clauses)[i];
    delete data;
    
    return good;
  }


 private:
  double max(const double& a, const double& b) { if (a > b) return a; return b;}

    // returns the index of the block which the ground predicate
    // with index predIndex is in. If not in any, returns -1
  int getBlock(const int& predIndex)
  {
    for (int i = 0; i < blocks_->size(); i++)
    {
      int block = (*blocks_)[i].find(predIndex);
      if (block >= 0)
        return i;
    }
    return -1;
  }

    // Sets the truth values for chain chain of all ground preds in the
    // block block except for the one with index predIndex
  void setOthersInBlockToFalse(const int& chain, const int& predIndex,
                               const int& blockIdx)
  {
    Array<int> block = (*blocks_)[blockIdx];
    for (int i = 0; i < block.size(); i++)
    {
      if (i != predIndex)
        (*gndPreds_)[block[i]]->setTruthValue(chain, false);
    }
  }
  
    // Returns index of chosen atom in block with index blockIdx
  int gibbsSampleFromBlock(const int& chain, const int& blockIdx)
  {
    Array<int> block = (*blocks_)[blockIdx];
    
    Array<double> numerators;
    double denominator = 0;
    
    for (int i = 0; i < block.size(); i++)
    {
      double prob = (*gndPreds_)[block[i]]->getProbability(chain);
      numerators.append(prob);
      denominator += prob;
    }
    float random = random_.random();

    double numSum = 0.0;    
    for (int i = 0; i < block.size(); i++)
    {
      numSum += numerators[i];
      if (random < (numSum / denominator))
        return i;
    }
    return block.size() - 1;
  }
  
  void addToLitWtArray(const int& lit, const double& wt, 
                       LitWtHashArray& litWtArr)
  {
    LitWt* litWt = new LitWt;
    litWt->lit = lit;
    litWt->wt  = wt;
    int a = litWtArr.find(litWt);
    if (a < 0)    litWtArr.append(litWt);
    else        { litWtArr[a]->wt += litWt->wt; delete litWt; } 
  }


  void storeFlippedGndPreds(const GroundClause* const & gndClause, 
                            const double& wt, LitWtHashArray& litWtArr)
  {
    int numGndPreds = gndClause->getNumGroundPredicates();
    double negWt = -wt/numGndPreds;
    for (int j = 0; j < numGndPreds; j++)
    {
      int idx = gndClause->getGroundPredicateIndex(j);
      int lit = (gndClause->getGroundPredicateSense(j)) ?
                //MaxWalksat::makeNegatedLiteral(idx) : idx;
                WSUtil::makeNegatedLiteral(idx) : idx;
      addToLitWtArray(lit, negWt, litWtArr);
    }
  }


    //The caller is responsible for deleting returned  maxWalksatData.
  MaxWalksatData* prepareMaxWalksatData()
  {  
      //this is a reasonable number to prevent the external executable 
      //MaxWalksat from have having overflow errors
    int maxAllowedWt = 4000;

      // prepare the propositional theory for use in MaxWalksat
    MaxWalksatData* data = new MaxWalksatData;
    data->numGndPreds = gndPreds_->size();

      //used to hold lits in unit clauses (including those formed by flipping 
      //negative wt clauses)
    LitWtHashArray litWtArr;
    Array<GroundClause*> negWtClauses;
    Array<GroundClause*> hardGndClauses; //used to store hard ground clauses

      // process ground clauses except the hard ones
    for (int i = 0; i < gndClauses_->size(); i++) 
    {
      GroundClause* gndClause = (*gndClauses_)[i];

      if (gndClause->isHardClause()){hardGndClauses.append(gndClause);continue;}

      assert(!gndClause->isHardClause());
      
        // Weight could be negative: The inference algorithms now
        // handle negative weights
      double wt = gndClause->getWt();

      int numGndPreds = gndClause->getNumGroundPredicates();
      if (numGndPreds == 1)
      {
        int idx = gndClause->getGroundPredicateIndex(0);
        int lit = (gndClause->getGroundPredicateSense(0)) ?
                  //idx : MaxWalksat::makeNegatedLiteral(idx);
                  idx : WSUtil::makeNegatedLiteral(idx);
        addToLitWtArray(lit, wt, litWtArr);
      }
      else
      {
        Array<int>* litClause = new Array<int>;
        litClause->growToSize(numGndPreds);
        for (int j = 0; j < numGndPreds; j++) 
        {
          int idx = gndClause->getGroundPredicateIndex(j);
          int lit = (gndClause->getGroundPredicateSense(j)) ?
                    //idx : MaxWalksat::makeNegatedLiteral(idx);
                    idx : WSUtil::makeNegatedLiteral(idx);
          (*litClause)[j]= lit;
        }
        data->clauses.append(litClause);
        data->clauseWts.append(wt);
        //cout << wt << " " << litClause << endl;
      }
    }

      // for each unit ground clause
    for (int i = 0; i < litWtArr.size(); i++)
    {
      Array<int>* litClause = new Array<int>;
      litClause->growToSize(1);
      (*litClause)[0]= litWtArr[i]->lit;
      data->clauses.append(litClause);
      double wt = litWtArr[i]->wt;
      data->clauseWts.append(wt);
      delete litWtArr[i];
    }

      // find minimum wt among soft clauses
    double minWt = DBL_MAX;
    double maxWt = DBL_MIN;
    for(int i = 0; i < data->clauseWts.size(); i++)
    {
    	//cout << (data->clauseWts)[i] << " " << (data->clauses)[i] << endl;
        // Weight could be negative
      double absWt = abs((data->clauseWts)[i]);
      if (absWt < minWt) minWt = absWt;
      else if (absWt > maxWt) maxWt = absWt;
      assert(minWt >= 0);
      assert(maxWt >= 0);
    }
    
      //find out how much to scale weights by
    double scale = 1;
    if (maxWt > maxAllowedWt) scale = maxAllowedWt/maxWt;
    else
    {
      if (minWt < 10)
      {
        scale = 10/minWt;
        if (scale*maxWt > maxAllowedWt) scale = maxAllowedWt/maxWt;
      }
    }

      // find the sum of all the weights of soft clauses, scaling up if needed
    int sumSoftWts = 0;
    for(int i = 0; i < data->clauseWts.size(); i++)
    {
      double oldwt = (data->clauseWts)[i];
      int newwt = (int)(oldwt*scale + 0.5);
      (data->clauseWts)[i] = (double)newwt;
      sumSoftWts += abs(newwt);
    }
    assert(sumSoftWts >= 0);
    
      // process the hard ground clauses
    int largeWt = (int)(sumSoftWts + 10*scale);
    data->hardClauseWt = (int)(largeWt/scale+0.5);
    for (int i = 0; i < hardGndClauses.size(); i++) 
    {
      GroundClause* gndClause = hardGndClauses[i];

      Array<int>* litClause = new Array<int>;
      litClause->growToSize(gndClause->getNumGroundPredicates());

      for (int j = 0; j < gndClause->getNumGroundPredicates(); j++) 
      {
        int idx = gndClause->getGroundPredicateIndex(j);
        int lit = (gndClause->getGroundPredicateSense(j)) ?
                  //idx : MaxWalksat::makeNegatedLiteral(idx);
                  idx : WSUtil::makeNegatedLiteral(idx);
        (*litClause)[j]= lit;
      }
      data->clauses.append(litClause);
      data->clauseWts.append(largeWt);
    }    

    return data;
  }

  void maxWalksatInitGndPredsTruthValues(const int& numChains, 
                                         const MaxWalksatParams* const & params)
  { 
    double hardClauseWt;
    for (int c = 0; c < numChains; c++)  
      runMaxWalksat(c, true, params, hardClauseWt); 
  }


  void addUnknownGndClauses(const GroundPredicate* const& queryGndPred,
                            Clause* const & c, const Domain* const & domain, 
                            const Database* const & db, 
                            const bool& genClauseForAllPredGndings,
                            const AddGroundClauseStruct* const & agcs)
  {
    
    if (genClauseForAllPredGndings)
      c->addUnknownClauses(domain, db, -1, NULL, agcs);
    else
    {
      for (int i = 0; i < c->getNumPredicates(); i++)
      {
        if (c->getPredicate(i)->canBeGroundedAs(queryGndPred))
          c->addUnknownClauses(domain, db, i, queryGndPred, agcs);
      }
    }
  } 


  void initConvergenceTests(GelmanConvergenceTest**& burnConvergenceTests,
                            ConvergenceTest**& gibbsConvergenceTests, 
                            const double& gamma, const double& epsilonFrac, 
                            const int& numGndPreds, const int& numChains)
  {
    burnConvergenceTests = new GelmanConvergenceTest*[numGndPreds];
    gibbsConvergenceTests = new ConvergenceTest*[numGndPreds];
    for (int i = 0; i < numGndPreds; i++) 
    {
      burnConvergenceTests[i]  = new GelmanConvergenceTest(numChains);
      gibbsConvergenceTests[i] = new ConvergenceTest(numChains, gamma,
                                                     epsilonFrac);
    }
  }

  
  void deleteConvergenceTests(GelmanConvergenceTest**& burnConvergenceTests,
                              ConvergenceTest**& gibbsConvergenceTests, 
                              const int& numGndPreds)
  {
    for (int i = 0; i < numGndPreds; i++) 
    {
      delete burnConvergenceTests[i];
      delete gibbsConvergenceTests[i];
    }
    delete [] burnConvergenceTests;
    delete [] gibbsConvergenceTests;
  }  


  bool genTruthValueForProb(const double& p)
  {
    if (p==1.0) return true;
    if (p==0.0) return false;
    return random_.random() <= p;    
  }


  void randomInitGndPredsTruthValues(const int& numChains)
  {
      // Randomizer
    int seed;
    struct timeval tv;
    struct timezone tzp;
    gettimeofday(&tv,&tzp);
    seed = (unsigned int)((( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec);
    srandom(seed);
    
    for (int c = 0; c < numChains; c++)
    {
        // For each block: select one to set to true
      for (int i = 0; i < blocks_->size(); i++)
      {
          // If evidence atom exists, then all others are false
        if ((*blockEvidence_)[i])
        {
            // If 2nd argument is -1, then all are set to false
          setOthersInBlockToFalse(c, -1, i);
          continue;
        }
        
        Array<int> block = (*blocks_)[i];
        int chosen = random() % block.size();
        (*gndPreds_)[block[chosen]]->setTruthValue(c, true);
        setOthersInBlockToFalse(c, chosen, i);
      }
      
        // Random tv for all not in blocks
      for (int i = 0; i < gndPreds_->size(); i++)
      {
          // Predicates in blocks have been handled above
        if (getBlock(i) == -1)
        {
          bool tv = genTruthValueForProb(0.5);
	      (*gndPreds_)[i]->setTruthValue(c, tv);
        }
      }
    }
  }

  void randomInitPerChainGndPredsTruthValues(const int& c)
  {
      for (int i = 0; i < gndPreds_->size(); i++)        
	    (*gndPreds_)[i]->setTruthValue(c, genTruthValueForProb(0.5));
  }
  
  void hardClauseInitGndPredsTruthValues(const int& numChains, SampleSat& sampleSat,
  										 int*& fixedAtoms, int**& clauses,
  										 bool*& assignments)
  {
  	cout << "Initialize to satisfy all hard clauses ..." << endl;
  	int numClauses = 0;
  	for (int i = 1; i <= gndPreds_->size(); i++) fixedAtoms[i] = -1;

      // generate sat query
  	for (int i = 0; i < gndClauses_->size(); i++)
  	{
      GroundClause* clause = (*gndClauses_)[i];
	  if ( clause->isHardClause() )
	  {
	     // unit clause
	  	if (clause->getNumGroundPredicates() == 1)
	  	{
          fixedAtoms[clause->getGroundPredicateIndex(0) + 1] =
          	clause->getGroundPredicateSense(0);
	  	}
      	else
      	{
          bool isSat = false;
          int k = 0;
          for (int j = 0; j < clause->getNumGroundPredicates(); j++)
          {
          	int idx = clause->getGroundPredicateIndex(j) + 1;
          	if (fixedAtoms[idx] == clause->getGroundPredicateSense(j))
          	{
              isSat = true;
              break;
		  	}
          	if (fixedAtoms[idx] >= 0) continue;	// fixed but not sat -> ignore
          	int lit = (clause->getGroundPredicateSense(j))? idx : (-idx);
          	clauses[numClauses][k++] = lit;
		  }
			// already sat, roll back
		  if (isSat) continue;
		  else clauses[numClauses++][k] = 0;
	  	}
	  }
  	}

 	  // init
  	for (int c = 0; c < numChains; c++)
  	{
	  	// sample for hard clauses
	  if (sampleSat.sample(assignments, numClauses, blocks_, blockEvidence_))
	  {
      	for (int i = 0; i < gndPreds_->size(); i++)
      	{
          (*gndPreds_)[i]->setTruthValue(c, assignments[i]);
	  	}
      }
      else
      {
	  	cout << "[" << c << " : numcl=" << numClauses
             << "] No solution found! Init randomly."<<endl;

          // For each block: select one to set to true
        for (int i = 0; i < blocks_->size(); i++)
        {
            // If evidence atom exists, then all others are false
          if ((*blockEvidence_)[i])
          {
              // If 2nd argument is -1, then all are set to false
            setOthersInBlockToFalse(c, -1, i);
            continue;
          }
        
          Array<int> block = (*blocks_)[i];
          int chosen = random() % block.size();
          (*gndPreds_)[block[chosen]]->setTruthValue(c, true);
          setOthersInBlockToFalse(c, chosen, i);
        }
      
          // Random tv for all not in blocks
        for (int i = 0; i < gndPreds_->size(); i++)
        {
            // Predicates in blocks have been handled above
          if (getBlock(i) == -1)
          {
            bool tv = genTruthValueForProb(0.5);
            (*gndPreds_)[i]->setTruthValue(c, tv);
          }
        }
	  }			  
  	}
  }

  void gndPredFlippedUpdates(const int& gndPredIdx, const int& chainIdx,
                             GroundPredicateHashArray& affectedGndPreds)
  {
    GroundPredicate* gndPred = (*gndPreds_)[gndPredIdx];
    affectedGndPreds.append(gndPred, gndPreds_->size());
    assert(affectedGndPreds.size() <= gndPreds_->size());

    const Array<GroundClause*>* gndClauses = gndPred->getGndClauses();
    const Array<bool>* senseInGndClauses = gndPred->getSenseInGndClauses();
    assert(gndClauses->size() == senseInGndClauses->size());
    GroundClause* gndClause;

    for (int i = 0; i < gndClauses->size(); i++)
    {
      gndClause = (*gndClauses)[i];
      if (gndPred->getTruthValue(chainIdx) == (*senseInGndClauses)[i])
        gndClause->incrementNumSatLiterals(chainIdx);
      else
        gndClause->decrementNumSatLiterals(chainIdx);

      assert(gndClause->getNumSatLiterals(chainIdx)
             <= gndClause->getNumGroundPredicates());

      for (int j = 0; j < gndClause->getNumGroundPredicates(); j++)
      {
        affectedGndPreds.append((GroundPredicate*)
                                 gndClause->getGroundPredicate(j), 
                                 gndPreds_->size());
        assert(affectedGndPreds.size() <= gndPreds_->size());
      }
    }
  }

  void updateClauses(const int& gndPredIdx, const int& chainIdx)
  {
    GroundPredicate* gndPred = (*gndPreds_)[gndPredIdx];
    const Array<GroundClause*>* gndClauses = gndPred->getGndClauses();
    const Array<bool>* senseInGndClauses = gndPred->getSenseInGndClauses();
    assert(gndClauses->size() == senseInGndClauses->size());
    GroundClause* gndClause;

    for (int i = 0; i < gndClauses->size(); i++)
    {
      gndClause = (*gndClauses)[i];
      if (gndPred->getTruthValue(chainIdx) == (*senseInGndClauses)[i])
        gndClause->incrementNumSatLiterals(chainIdx);
      else
        gndClause->decrementNumSatLiterals(chainIdx);

      assert(gndClause->getNumSatLiterals(chainIdx)
             <= gndClause->getNumGroundPredicates());
    }
  }


  void updateWtsForAffectedGndPreds(GroundPredicateHashArray& affectedGndPreds,
                                    const int& chainIdx)
  {
   
      // for each ground predicate whose MB has changed
    for (int g = 0; g < affectedGndPreds.size(); g++)
    {

      GroundPredicate* gndPred = affectedGndPreds[g];
      double wtIfNoChange = 0, wtIfInverted = 0, wt;
      const Array<GroundClause*>* gndClauses = gndPred->getGndClauses();
      const Array<bool>* senseInGndClauses = gndPred->getSenseInGndClauses();
      assert(gndClauses->size() == senseInGndClauses->size());

      for (int i = 0; i < gndClauses->size(); i++)
      {      
        wt = (*gndClauses)[i]->getWt();
	
        if ((*gndClauses)[i]->getNumSatLiterals(chainIdx) > 1) 
        {
            // some other literal is making it sat, so it doesn't matter
          wtIfNoChange += wt;
          wtIfInverted += wt;
        }
        else
        if ((*gndClauses)[i]->getNumSatLiterals(chainIdx) == 1) 
        {
          wtIfNoChange += wt;
            // if the current truth value is the same as its sense in gndClause
          if (gndPred->getTruthValue(chainIdx) == (*senseInGndClauses)[i]) 
          {
            // This gndPred is the only one making this function satisfied
          }
          else 
          {
              // Some other literal is making it satisfied
            wtIfInverted += wt;
          }
        }
        else
        if ((*gndClauses)[i]->getNumSatLiterals(chainIdx) == 0) 
        {
          // None satisfy, so when gndPred switch to its negative, it'll satisfy
          wtIfInverted += wt;
        }
      } // for each ground clause that gndPred appears in


      if (gndPred->getTruthValue(chainIdx)) 
      {
        gndPred->setWtWhenTrue(chainIdx, wtIfNoChange);
        gndPred->setWtWhenFalse(chainIdx, wtIfInverted);
      }
      else 
      {
        gndPred->setWtWhenFalse(chainIdx, wtIfNoChange);
        gndPred->setWtWhenTrue(chainIdx, wtIfInverted);
      }
    } // for each ground predicate whose MB has changed
  }

  // --- Calculate energy as the sum of weights of sat clauses
  double getWeightSum(int ci)
  {
    double w = 0;
    for (int i=0; i<gndClauses_->size(); i++)
    {
      GroundClause* clause = (*gndClauses_)[i];
      double wt = clause->getWt();
      if (clause->getNumSatLiterals(ci)>0) w+=wt;
	}
	return w;
  }

 public:

  
    // initialize the structure to store the number of satisfied literals in
    // each clause
  void initNumSatLiterals(const int& numChains)
  { 
    for(int i=0;i<gndClauses_->size();i++)
    {
      (*gndClauses_)[i]->initNumSatLiterals(numChains);
    }
  }
  
    // get the number of clause gndings using the first order clause frequencies
  void getNumClauseGndings(double numGndings[], int clauseCnt, bool tv,
                           int chainIdx, const Domain *domain)
  {
    IntPairItr itr;
    IntPair *clauseFrequencies;
	 
    for(int clauseno = 0; clauseno < clauseCnt; clauseno++)
      numGndings[clauseno] = 0;
    
    for(int i = 0; i < gndClauses_->size(); i++)
    {
      GroundClause *gndClause = (*gndClauses_)[i];
      int satLitcnt = gndClause->getNumSatLiterals(chainIdx);
      if(tv == true && satLitcnt == 0)
        continue;
      if(tv == false && satLitcnt > 0)
        continue;
		  
      clauseFrequencies = gndClause->getClauseFrequencies();
      for(itr = clauseFrequencies->begin(); itr != clauseFrequencies->end(); itr++)
      {
        int clauseno = itr->first;
        int frequency = itr->second;
        numGndings[clauseno] += frequency;
      }
    }
  }

 private:
  Array<GroundPredicate*>* gndPreds_;
  Array<GroundClause*>* gndClauses_;
  Random random_;
    // Blocks of gndPred indices which belong together
  Array<Array<int> >* blocks_;
    // Flags indicating if block is fulfilled by evidence
  Array<bool>* blockEvidence_;


  Array<ClauseWithInfo*>* ungndClausesWithInfo_;
  Array<Array<int>*>* predAppearsInGndClauses_; //each int is the index of the gnd clause
  Array<Array<int>*>* ungndClausePointers_; //for each gnd clause an array of ungndclauses
};


#endif
