#include "clause.h"
#include "ut-mrf.h"

int Clause::addUnknownClauses_ut(const Domain* const & domain, 
				 const Database* const& db, const int& gndPredIdx, 
				 const GroundPredicate* const & groundPred,
				 const AddGroundClauseStruct* const & agcs, UTMRF *mrf,
				 int ungndClauseId){
  createVarIdToVarsGroundedType(domain);
  if (gndPredIdx >= 0) groundPredVars(gndPredIdx, groundPred);
  int toReturn = (int) countNumTrueGroundings_ut(domain, db, true, false, gndPredIdx, groundPred, 
						 NULL, NULL, NULL, NULL, agcs, mrf, ungndClauseId);
  restoreVars();
  deleteVarIdToVarsGroundedType();
  return toReturn;
}


bool Clause::createAndAddUnknownClause_ut(
                                 Array<GroundClause*>* const& unknownGndClauses,
                                 Array<Clause*>* const& unknownClauses,
                                 double* const & numUnknownClauses,
                                 const AddGroundClauseStruct* const & agcs, UTMRF *mrf,
				 int ungndClauseId)
{ 
  PredicateSet predSet; // used to detect duplicates
  PredicateSet::iterator iter;
  
  Clause* clause = NULL;
  for (int i = 0; i < predicates_->size(); i++)
  {
    Predicate* predicate = (*predicates_)[i];
    assert(predicate->isGrounded());
    if (predicate->getTruthValue() == UNKNOWN)
    {
      if ( (iter=predSet.find(predicate)) != predSet.end() )
      {
          // the two gnd preds are of opp sense, so clause must be satisfied
        if ((*iter)->getSense() !=  predicate->getSense())
        {
          if (clause) delete clause;
          return true;
        }
        // since the two gnd preds are identical, no point adding a dup
        continue;
      }
      else
        predSet.insert(predicate);
      
      if (clause == NULL) clause = new Clause();
      Predicate* pred = new Predicate(*predicate, clause);
      clause->appendPredicate(pred);
    }
  }
  
  if (clause) 
  {
    if (numUnknownClauses) (*numUnknownClauses)++;

    clause->setWt(wt_);
    clause->canonicalizeWithoutVariables();
    //cout << "unknown clause: "; clause->print(cout); cout << endl;
    //cout << "created from "; print(cout); cout << endl;
    
    if (agcs) mrf->addUnknownGndClause(agcs, this, clause, isHardClause_, ungndClauseId);
    
    if (unknownGndClauses)
    {
      unknownGndClauses->append(new GroundClause(clause));
      if (isHardClause_) unknownGndClauses->lastItem()->setWtToHardWt();
    }
    else
    if (unknownClauses)
    {
      unknownClauses->append(clause);
      if (isHardClause_) clause->setIsHardClause(true);
    }
    if (unknownClauses == NULL) delete clause;
  }
  
  return false;
}


double Clause::countNumTrueGroundings_ut(const Domain* const & domain,
					 const Database* const & db,
					 const bool& hasUnknownPreds,
					 const bool& checkSatOnly,
					 //params below: find unknown clauses
					 const int& gndPredIdx,
					 const GroundPredicate* const & groundPred,
					 const Predicate* const & gndPred,
					 Array<GroundClause*>* const & unknownGndClauses,
					 Array<Clause*>* const & unknownClauses,
					 double* const & numUnknownClauses,
					 //params below: add unknown clauses to MRF
					 const AddGroundClauseStruct* const & agcs, UTMRF *mrf,
					 int ungndClauseId)
  {
    assert(unknownGndClauses == NULL || unknownClauses == NULL);
    assert(groundPred == NULL || gndPred == NULL);

    if (numUnknownClauses) *numUnknownClauses = 0;

    bool findUnknownClauses = (unknownGndClauses || unknownClauses || 
                               numUnknownClauses || agcs);
      // these predicates must not be grounded as groundPred/gndPred
    Array<Predicate*> bannedPreds;
    if (findUnknownClauses)
      getBannedPreds(bannedPreds, gndPredIdx, groundPred, gndPred);

    unsigned int numTrueGndings = 0;
    
      //Copy the literals so that their original order in the clause is
      //not affected by the subsequent sorting
    Array<Predicate*> clauseLits(*predicates_);

      //Sort preds in decreasing order of #TrueGndOfLiteral/#numOfGroundings.
      //The larger the number of true groundings of a literal, the more likely
      //it is to be true, so put it in front so that we can decide whether the
      //clause is true early.The larger the number of groundings of the literal,
      //the larger the savings when we decide that preceding literals are true.
    sortLiteralsByTrueDivTotalGroundings(clauseLits, domain, db);

      //simulate a stack, back/front corresponds to top/bottom
      //ivg stands for index, varIds, groundings
    Array<LitIdxVarIdsGndings*> ivgArr;
    createAllLitIdxVarsGndings(clauseLits, ivgArr, domain);
    if (findUnknownClauses) createBannedPreds(clauseLits, ivgArr, bannedPreds);
    int ivgArrIdx = 0; //store current position in ivgArr
    bool lookAtNextLit = false;
    
      // while stack is not empty
    while (ivgArrIdx >= 0)
    {
        //get variable groundings at top of stack
      LitIdxVarIdsGndings* ivg = ivgArr[ivgArrIdx];
      Predicate* lit = clauseLits[ivg->litIdx];
      Array<int>& varIds = ivg->varIds;
      ArraysAccessor<int>& varGndings = ivg->varGndings;
      bool& litUnseen = ivg->litUnseen;
      bool hasComb;

        // while there are groundings of literal's variables
      while((hasComb=varGndings.hasNextCombination()) || litUnseen)
      {
          // there may be no combinations if the literal is fully grounded
        if (litUnseen) litUnseen = false;

        if (hasComb)
        {
            //ground the literal's variables throughout the clause
          int constId;
          int v = 0; // index of varIds
            //for each variable in literal
          while(varGndings.nextItemInCombination(constId))
          {
            Array<Term*>& vars =(*varIdToVarsGroundedType_)[-varIds[v++]]->vars;
            for (int i = 0; i < vars.size(); i++) vars[i]->setId(constId);
          }
          
          //return values of grounded functions are assigned to their parent
          //terms in literalOrSubsequentLiteralAreTrue()
        }

          //if literal or subsequent grounded literals are true,
        if (literalOrSubsequentLiteralsAreTrue(lit, ivg->subseqGndLits, db))
        {
          if (checkSatOnly) return 1;
            //count the number of combinations of remaining variables
          int numComb = 1;
          for (int i = ivgArrIdx+1; i < ivgArr.size(); i++)
          {
            int numVar = ivgArr[i]->varGndings.getNumArrays();
            for (int j = 0; j < numVar; j++)
              numComb *= ivgArr[i]->varGndings.getArray(j)->size();
          }
          numTrueGndings += numComb;
        }
        else
        if (findUnknownClauses && 
            bannedPredsAreGndedAsGndPred(ivg->bannedPreds, groundPred, gndPred))
        {
          //do nothing, will move down stack later
        }
        else
        {
            // if there are more literals
          if (ivgArrIdx+1 < ivgArr.size())
          {
            lookAtNextLit = true;
            ivgArrIdx++; // move up stack
            break;
          }
            //At this point all the literals are grounded, and they are either
            //unknown or false (have truth values opposite of their senses).

          //cout << "false or unknown clause:\n\t";
          //printWithoutWtWithStrVar(cout, domain); cout << endl;

          bool twoLitWithOppSense = false;
          if (hasUnknownPreds)
          {
            if (hasTwoLiteralsWithOppSense()) 
            {
              twoLitWithOppSense = true;
              ++numTrueGndings;
              if (checkSatOnly) return 1;
            }
          }

          if (!twoLitWithOppSense && findUnknownClauses)
          {
            assert(!containsGndPredBeforeIdx(gndPredIdx, groundPred, gndPred));

              //Create a new clause by appending unknown predicates to it.
            createAndAddUnknownClause_ut(unknownGndClauses, unknownClauses, 
					 numUnknownClauses, agcs, mrf, ungndClauseId);
          }
        }
      } //while there are groundings of literal's variables

        //if we exit the while loop in order to look at next literal 
        //(i.e. without considering all groundings of current literal)
      if (lookAtNextLit) { lookAtNextLit = false; }
      else { varGndings.reset(); litUnseen = true; ivgArrIdx--; }//mv down stack

    } // while stack is not empty

    deleteAllLitIdxVarsGndings(ivgArr);

    return numTrueGndings;
  }
