#ifndef PREDICATESMAPPING_H_
#define PREDICATESMAPPING_H_

/*
 * Author: Tuyen Huynh
 * Description: Find the best predicate mapping between the source and target domain.
 */

#include <fstream>
#include <iostream>
#include <cctype>
//#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <sys/dir.h>
#include <stdlib.h>
#include <set.h>
#include "arguments.h"
#include "fol.h"
#include "learnwts.h"
#include "infer.h"
#include "StringTokenizer.h"
#include "array.h"
#include "hasharray.h"
#include "equalstr.h"
#include "util.h"
#include "wpll.h"
#include "ut-structlearn.h"

class EqualPredicateTemplate
{
 public:
  bool operator()(const PredicateTemplate* const & p1, const PredicateTemplate* const & p2) const
  {
  	if (strcmp(p1->getName(),p2->getName()) != 0) return false;
  	if (p1->getNumTerms() != p2->getNumTerms()) return false;
  	for (int i =0; i < p1->getNumTerms(); i++) {
  		if (strcmp(p1->getTermTypeAsStr(i),p2->getTermTypeAsStr(i)) != 0)
  			return false;
  	}
    return true;
  }
};

class HashPredicateTemplate
{
 public:
  size_t operator()(const PredicateTemplate* const & p) const
  {
    return hash<char const *>()(p->getName());
  }
};

typedef hash_map<const char*, const StringHashArray* , hash<const char*>, EqualStr> TypeMap;
typedef hash_map<const char*, const char* , hash<const char*>, EqualStr> StringMap;
typedef hash_map<const PredicateTemplate*,const PredicateTemplate*, HashPredicateTemplate, EqualPredicateTemplate> PredicateMap;
typedef HashArray<const PredicateTemplate*, HashPredicateTemplate, EqualPredicateTemplate> PredicateTemplateSet;

class PredicatesMapping
{
 public:
 	PredicatesMapping(char* _sourceMLNFile, char* _targetPredicateFile, char* _targetDbFiles, 
 						char* _mappingConstraintFile, char* _outTargetMLNFile, 
 						bool& _globalMapping, bool& _sampling, bool& _debug, bool& _doRevision, double _confidentRatio)	
 	: sourceMLNFile(_sourceMLNFile), targetPredicateFile(_targetPredicateFile), targetDbFiles(_targetDbFiles),
 	  mappingConstraintFile(_mappingConstraintFile), outTargetMLNFile(_outTargetMLNFile), 
 	  globalMapping(_globalMapping), sampling(_sampling), debug (_debug), doRevision(_doRevision), confidentRatio(_confidentRatio),
 	  mappingIndex(1), bestWPLLBeforeRevising(-1E+10), bestWPLLAfterRevising(-1E+10), bestMappingIndex(0)
    {
    	// setup temporary files
    	tmpSrcMLNFile = "tmp-";
    	tmpSrcMLNFile.append(sourceMLNFile);
    	tmpSrcMLNFile.append("-");
    	tmpSrcMLNFile.append(outTargetMLNFile);
    	
    	tmpTargetPredicateFile = "tmpTargetPredicate-";
    	tmpTargetPredicateFile.append(outTargetMLNFile);
    	
    	tmpPredMappingFile = "tmpPredMapping-";
    	tmpPredMappingFile.append(outTargetMLNFile);
    	
    	tmpTargetMLNFile = "tmpTargetMLN-";
    	tmpTargetMLNFile.append(outTargetMLNFile);
    	
    	bestPredMappingFile = "bestPredMapping-";
    	bestPredMappingFile.append(outTargetMLNFile);
    	
    	logFile = "";
    	logFile.append(outTargetMLNFile);
    	
    	tmpPredMappingFile.append(".txt");
    	tmpTargetMLNFile.append(".mln");
    	bestPredMappingFile.append(".txt");
    	logFile.append(".log");
    }
 
 
	 void run() 
	 {		
	  Timer timer;
  	  double startSec = timer.time();  	  
  	  
	    //Compute the size in MB of the components of Term, Predicate, and Clause
	    //that do not change. The sizes will be used when computing the running 
	    //size of the cache of clauses.
	  Term::computeFixedSizeB();
	  Predicate::computeFixedSizeB();
	  Clause::computeFixedSizeB();
	  AuxClauseData::computeFixedSizeB();
	  
	  ///////////////////////// check and extract the parameters //////////////////
	  
	  //the second .mln file to the last one in inMLNFiles _may_ be used 
	  //to hold constants, so they are held in constFilesArr. They will be
	  //included into the first .mln file.
	
	
	  ////////////////////////// create source and target domains and mlns ///////////////////////////
	  Domain* sourceDomain = new Domain;;
	  MLN* sourceMLN = new MLN();
	  
	  bool addUnitClauses = false;
	  bool warnAboutDupGndPreds = true;
	  bool mustHaveWtOrFullStop = true;
	  bool flipWtsOfFlippedClause = false;
	  bool allPredsExceptQueriesAreCW = false;
	  bool ok = false;
	
	  double begSec = timer.time();
	  cout << "Parsing MLN and creating source domains..." << endl;
	  ok = runYYParser1(sourceMLN, sourceDomain, sourceMLNFile, allPredsExceptQueriesAreCW, 
	                   NULL, NULL, addUnitClauses, 
	                   warnAboutDupGndPreds, 0, mustHaveWtOrFullStop, 
	                   NULL, true, flipWtsOfFlippedClause);
	  if (!ok) {
	  	cout << "Parsing source MLN failed" << endl;
	  	exit(1);
	  }
	  cout << "Parsing MLN and creating source domains took ";
	  Timer::printTime(cout, timer.time()-begSec); cout << endl << endl;
	  
	  // output the source MLN without predicate declaration
	  ofstream out;
	  out.open(tmpSrcMLNFile.c_str());
	  printMLNWithoutPredDeclaration(out, sourceMLN, sourceDomain);
	  out.close();
	  
	  //PredicateTemplateSet sourcePredicates;
	  sourcePredicates.clear();
	  for (int i = 0; i < sourceDomain->getNumPredicates(); i++) {
	  	const PredicateTemplate* currentPred = sourceDomain->getPredicateTemplate(i);
	  	if (!(currentPred->isEqualPredWithType()))
	  		sourcePredicates.append(currentPred);
	  }
	  
	  if (debug) printPredicates(sourcePredicates);
	  
	  Domain* targetDomain = new Domain;
	  MLN* targetMLN = new MLN();
	  begSec = timer.time();
	  cout << "Parsing MLN and creating target domains..." << endl;
	  ok = runYYParser1(targetMLN, targetDomain, targetPredicateFile, allPredsExceptQueriesAreCW, 
	                   NULL, NULL, addUnitClauses, 
	                   warnAboutDupGndPreds, 0, mustHaveWtOrFullStop, 
	                   NULL, true, flipWtsOfFlippedClause);
	  cout << "Parsing MLN and creating target domains took ";
	  Timer::printTime(cout, timer.time()-begSec); cout << endl << endl;
	  
	  if (!ok) {
	  	cout << "Parsing target MLN failed" << endl;
	  	exit(1);
	  }
	  
	  //PredicateTemplateSet targetPredicates;
	  targetPredicates.clear();
	  for (int i = 0; i < targetDomain->getNumPredicates(); i++) {
	  	const PredicateTemplate* currentPred = targetDomain->getPredicateTemplate(i);
	  	if (!(currentPred->isEqualPredWithType()))
	  		targetPredicates.append(currentPred);
	  }
	  
	  if (debug) printPredicates(targetPredicates);
	  
	
	  ////////////////////////////// parsing type constraints ///////////////////////////////
	  StringMap* typeConstraints = new StringMap[2];
	  /*
	  if (mappingConstraintFile != NULL) {
	  	cout << "Parsing type constraint file ..." << endl;
		string line;
	  	ifstream constraintFile(mappingConstraintFile);
	  	if (constraintFile.is_open())
	  	{
	    	while (! constraintFile.eof() )
	    	{
	      		getline (constraintFile,line);
	      		if (debug) cout << line << endl;
		    	StringTokenizer tokenizer = StringTokenizer(line,"->");
		    	if (tokenizer.countTokens() > 1) {
		    		// get the source type
		    		std::string srcTypeStr = tokenizer.nextToken();
		    		trim2(srcTypeStr);
		    		char* srcType = new char[strlen(srcTypeStr.c_str())+1];
	    			strcpy(srcType, srcTypeStr.c_str());
		    		if (debug)
		    			cout << "Source type: " << srcType << ";";
		    		// get the target types
		    		if (debug)
		    			cout << "Target type: ";
		    		StringHashArray* targetTypes = new StringHashArray;
		    		while (tokenizer.hasMoreTokens()) {
		    			std::string targetTypeStr = tokenizer.nextToken(",");
		    			trim2(targetTypeStr);
		    			char* targetType = new char[strlen(targetTypeStr.c_str())+1];
	    				strcpy(targetType, targetTypeStr.c_str());
		    			if (debug) cout << targetType << ",";
		    			targetTypes->append(targetType);
		    		}
		    		typeConstraints[srcType] = targetTypes;
		    		if (debug) cout << endl;
		    	}
	    	}
	    	constraintFile.close();
	  	}
	  	else cout << "Unable to open file: " << mappingConstraintFile << endl;
	  }
	  cout << "Finish parsing constraint file" << endl;
	 
	  if (debug) printTypeConstraints(typeConstraints); */
	  
	  /////////////////////////////// get target database files /////////////////////////////
	  /*
	  struct direct **files;
	  std::string* tmpFilename;
	  targetDbFilesArr = new Array<const char*>;
	  int count = scandir(targetDbDir, &files, file_select, alphasort);
	  for (int i=0; i<count; ++i) {
	  	tmpFilename = new string(targetDbDir);
	  	tmpFilename->append("/");
	  	tmpFilename->append(files[i]->d_name);
		targetDbFilesArr->append(tmpFilename->c_str());
	  }
	  */
	  
	  ////////////////////////////// predicates mapping ////////////////////////////////////
	  logOutput.open(logFile.c_str());
  	  bestTargetMLNOutput.open(outTargetMLNFile);
	  bestPredMappingOutput.open(bestPredMappingFile.c_str());
	  std::string tmpString(targetPredicateFile);
	  // copy target predicates to target output file
	  copyFromFile(tmpString, bestTargetMLNOutput);
	  // copy targetPredicateFile to tmpTargetPredicateFile
	  out.open(tmpTargetPredicateFile.c_str());
	  copyFromFile(tmpString, out);
	  out.close();
	  
	  if (globalMapping) {
	   	PredicateMap predicatesMapping;
	   	PredicateMap bestPredicateMapping;
	   	globalPredicateMapping(sourcePredicates, targetPredicates, 0, predicatesMapping, typeConstraints, bestPredicateMapping);
	   	logOutput << "Best mapping is mapping " << bestMappingIndex << " with average WPLL = " << bestWPLLAfterRevising << endl;
		printPredMapping(bestPredMappingOutput, bestPredicateMapping);
		// generate the best mapping file
		std::string tmpFile("tmp-");
		tmpFile.append(outTargetMLNFile);
		tmpFile.append("-global");
		std::string command = "sourceMLNToTargetMLN.pl " + tmpSrcMLNFile + " " + bestPredMappingFile + " " + tmpFile;
		//system(command.c_str());
		sourceMLNToTargetMLN(tmpSrcMLNFile, bestPredicateMapping, tmpFile);
		
		copyFromFile(tmpFile, bestTargetMLNOutput);
		command = "rm ";
		command.append(tmpFile);
		system(command.c_str());
	  }	
	  else
		localPredicateMapping(sourceMLN, sourceDomain, sourcePredicates, targetMLN, targetDomain, targetPredicates, typeConstraints);
	  
	  logOutput << "Total time take = ";
	  printTime1(logOutput, timer.time()-startSec);
	  
	  bestPredMappingOutput.close();
	  bestTargetMLNOutput.close();
	  logOutput.close();
	  
	  // remove temporary file
	  std::string command = "rm " + tmpSrcMLNFile;
	  system(command.c_str());
	  command = "rm " + tmpTargetPredicateFile;
	  system(command.c_str());
	 }
 
 private:
 
	bool isCompatible(const PredicateTemplate* srcPred, const PredicateTemplate* targetPred, StringMap typeConstraints[]){
		if (debug) {
			logOutput << "running isCompatible()" << endl;
			logOutput << "Source predicate: " << (*srcPred) << endl;
			logOutput << "Target predicate: " << (*targetPred) << endl;
			logOutput << "Source to Target type constraints: " << endl;
			printTypeConstraints(logOutput, typeConstraints[0]);
			logOutput << "Target to Source type constraints: " << endl;
			printTypeConstraints(logOutput, typeConstraints[1]);
		}

		if (srcPred->getNumTerms() != targetPred->getNumTerms()) {
			return false;
		}
		
		StringMap predicateTypeMap[2];
		for (int j = 0; j < srcPred->getNumTerms(); j++) {
			const char* srcType = srcPred->getTermTypeAsStr(j);
			const char* targetType = targetPred->getTermTypeAsStr(j);
			const char* targetTypeConstraint = predicateTypeMap[0][srcType];
			const char* srcTypeConstraint = predicateTypeMap[1][targetType];
			if ((targetTypeConstraint == NULL) &&
				(srcTypeConstraint == NULL)) 
			{
				predicateTypeMap[0][srcType] = targetType;
				predicateTypeMap[1][targetType] = srcType;
			}
			else {
				if ((targetTypeConstraint == NULL) ||
					 (srcTypeConstraint == NULL))
					 return false;
				if ((strcmp(targetTypeConstraint, targetType) != 0) ||
				    (strcmp(srcTypeConstraint, srcType) != 0))
					return false;
			}
		}
		
		//if (!(typeConstraints[0].empty())) {
			for (int i = 0; i < srcPred->getNumTerms(); i++) {
				const char* srcType = srcPred->getTermTypeAsStr(i);
				const char* targetType = targetPred->getTermTypeAsStr(i);
				const char* targetTypeConstraint = typeConstraints[0][srcType];
				const char* srcTypeConstraint = typeConstraints[1][targetType];
				if ((targetTypeConstraint != NULL) ||
					(srcTypeConstraint != NULL))
				{ 
					if ((targetTypeConstraint == NULL) ||
					 	(srcTypeConstraint == NULL))
					 	return false;
					if ((strcmp(targetTypeConstraint, targetType) != 0) ||
						 (strcmp(srcTypeConstraint, srcType) != 0))
						return false;
				}
			}
		//}
		return true;
	}

	double computeWPLL(const char* mlnFiles, const char* dbFiles,const char* outMLNFile, bool doRevision) {
	  //Compute the size in MB of the components of Term, Predicate, and Clause
	  //that do not change. The sizes will be used when computing the running 
	  //size of the cache of clauses.
	  Term::computeFixedSizeB();
	  Predicate::computeFixedSizeB();
	  Clause::computeFixedSizeB();
	  AuxClauseData::computeFixedSizeB();
	   
	  //the second .mln file to the last one in inMLNFiles _may_ be used 
	  //to hold constants, so they are held in constFilesArr. They will be
	  //included into the first .mln file.
	    //extract .mln and .db, file names
	  Array<string> constFilesArr, dbFilesArr, funcFilesArr;
	  extractFileNames(mlnFiles, constFilesArr);
	  assert(constFilesArr.size() >= 1);
	  string inMLNFile = constFilesArr[0];
	  constFilesArr.removeItem(0);
	  extractFileNames(dbFiles, dbFilesArr);
	  //extractFileNames(funcFiles, funcFilesArr);
		
	  ////////////////////////// create domains and mlns ///////////////////////////
	  bool multipleDatabases = true;
	  double ddelta = 0.05;
	  double epsilon = 0.2;
	  int minClauseSamples = -1;
	  int maxClauseSamples = -1;
		
	  double fraction = 0.8;
	  int minGndPredSamples = -1;
	  int maxGndPredSamples = -1;
		
	  bool noPrior = true;
	  double priorMean = 0;
	  double priorStdDev = 100;
		
	  int lbMaxIter = 10000;
	  double lbConvThresh = 1e-5;
	  int looseMaxIter = 10;
	  double looseConvThresh = 1e-3;
	  int cacheSize = 500;
	  bool noWtPredsEqually = false;
	  bool addUnitClauses = false;
	  
	  //Lily's parameters
	  int beamSize = 5;
	  double minWt = 0.0001;
	  double penalty = 0.01;
	  int maxVars = 6;
		
	  bool noSampleClauses = false;
	  bool noSampleGndPreds = false;
				
	  int numEstBestClauses = 10;
	  bool startFromEmptyMLN = false;
	  bool tryAllFlips = false;
	  int  bestGainUnchangedLimit = 2;
	  bool reEvalBestCandidatesWithTightParams = true;
	  
	  double percentageVotes = 0.1;
	  int numIters = 1;
	  
	  Array<Domain*> domains;
	  Array<MLN*> mlns;
	  Array<string> nonEvidPredNames;
	  StringHashArray* queryPredNames = NULL;
	    
	  createDomainsAndMLNs(domains, mlns, multipleDatabases, inMLNFile, 
	                       constFilesArr, funcFilesArr, dbFilesArr, queryPredNames,
	                       addUnitClauses, priorMean, false);
	
	  if (nonEvidPredNames.size() == 0) 
	    domains[0]->getNonEqualPredicateNames(nonEvidPredNames);
	  bool cacheClauses = (cacheSize > 0);
	  
	  if (!doRevision) {
		  WPLL wpll(&mlns, &domains, 
		                 &nonEvidPredNames, cacheClauses,cacheSize,
		                 sampling, ddelta, epsilon, minClauseSamples, maxClauseSamples,
		                 !noPrior, priorMean, priorStdDev, 
		                 !noWtPredsEqually, 
		                 lbMaxIter, lbConvThresh, looseMaxIter, looseConvThresh, 
		                 sampling,fraction,minGndPredSamples,maxGndPredSamples);
		  wpll.run();
		
		  ////////////////////////////// clean up ////////////////////////////////////
		  deleteDomains(domains);
		  for (int i = 0; i < mlns.size(); i++) delete mlns[i];
		  return wpll.getScore();	
	  }
	  else {
	  	logOutput << "doRevision ..." << endl;
	  	UtStructLearn sl(&mlns, startFromEmptyMLN, outMLNFile, &domains, 
		     &nonEvidPredNames, maxVars, cacheClauses,cacheSize,tryAllFlips,
		     !noSampleClauses, ddelta, epsilon, 
		     minClauseSamples, maxClauseSamples,
		     !noPrior, priorMean, priorStdDev, 
		     !noWtPredsEqually, 
		     lbMaxIter, lbConvThresh, looseMaxIter, looseConvThresh, 
		     beamSize, bestGainUnchangedLimit, numEstBestClauses, 
		     minWt, penalty, 
		     !noSampleGndPreds,fraction,minGndPredSamples,maxGndPredSamples,
		     reEvalBestCandidatesWithTightParams);
	  	logOutput << "calling reviseOneClause ..." << endl;
		double score = sl.reviseOneClause(numIters, percentageVotes);
		
		////////////////////////////// clean up ////////////////////////////////////
		deleteDomains(domains);
		for (int i = 0; i < mlns.size(); i++) delete mlns[i];

		return score;
	  }
	}// computeWPLL

	char * toUppercase(char * str) {
		int pos = 0;
		while (str[pos] != '\0') {
			str[pos] = toupper(str[pos]);
			pos++;
		}
		return str;
	}
	
	char * toLowercase(char * str) {
		int pos = 0;
		while (str[pos] != '\0') {
			str[pos] = tolower(str[pos]);
			pos++;
		}
		return str;
	}

	void replace(string& origin, string& pattern, string& replacement){
		while (origin.find(pattern)!= string::npos) {
			if (replacement == "") {
				origin = "";
				return;
			}	
			origin.replace(origin.find(pattern),pattern.length(),replacement);
		}
	}

	void sourceMLNToTargetMLN(string& _sourceMLNFile, PredicateMap& predsMapping, string& _targetMLNFile) {
		// read predicate mapping
		StringMap tmpPredicateMap;
  		set<const char *> targetPreds;
  		PredicateMap::iterator it;
		for (it=predsMapping.begin(); it!= predsMapping.end(); it++) {
			const PredicateTemplate* currentSrcPred = (*it).first;
			const char* srcPredStr = currentSrcPred->getName();
			if (debug)
				cout << "srcPred = " << srcPredStr << "; ";
			const PredicateTemplate* currentTargetPred = (*it).second;
			const char* targetPredStr = currentTargetPred->getName();
			if (debug)
				cout << "tarPred = " << targetPredStr << endl;
			if (strcmp(targetPredStr,"") != 0) {
				targetPreds.insert(targetPredStr);
				char * tmp = new char[80];
				strcpy(tmp, "911");
				strcat(tmp,targetPredStr);
				strcat(tmp,"911");
				toUppercase(tmp);
				tmpPredicateMap[srcPredStr] = tmp;
			}
			else
				tmpPredicateMap[srcPredStr] = "";
		}
		
		// reading source MLN file
		ifstream srcMLNFile(_sourceMLNFile.c_str());
		ofstream targetMLNFile(_targetMLNFile.c_str());
		std::string line;
		if (srcMLNFile.is_open()) {
			while (! srcMLNFile.eof()) {
				getline(srcMLNFile, line);
				trim2(line);
				if (strcmp(line.c_str(),"\r") != 0) {
					if (debug) cout << "Source line: " << line.c_str() << endl;
					
					// translate into internal representation
					for (int i = 0; i < sourcePredicates.size(); i++) {
						const PredicateTemplate* srcPred = sourcePredicates[i];
						std::string tmpSrcPred(srcPred->getName()); 
						if (debug) cout << "srcPred = " << tmpSrcPred.c_str() << "; ";
						std::string tmpTargetPred;
						const char* tmp = tmpPredicateMap[srcPred->getName()];
						if (tmp == NULL)
							tmpTargetPred = "";
						else
							tmpTargetPred = tmp;
						
						if (debug) cout << "tarPred = " << tmpTargetPred.c_str() << endl;
						replace(line,tmpSrcPred,tmpTargetPred);
					}
					if (debug)
						cout << "Temporary converted line : " << line << endl;
						
					// translate into target domain
					if (line != "") {
						set<const char *>::iterator iter;
						for (iter = targetPreds.begin(); iter != targetPreds.end(); iter++) {
							std::string replacement = (*iter);
							/*
							char * tmp = new char[replacement.length()];
		    				strcpy(tmp, replacement.c_str());
		    				toUppercase(tmp);
							*/
							char tmp[80];
							strcpy(tmp, "911");
							strcat(tmp, replacement.c_str());
							strcat(tmp,"911");
							toUppercase(tmp);
							std::string pattern(tmp);
							replace(line,pattern,replacement);
						}
					}
					if (debug)
						cout << "Converted line : " << line << endl;
					targetMLNFile << line << endl;
				}
			}
			srcMLNFile.close();
			targetMLNFile.close();
		}
		else cout << "Unable to open file: " << sourceMLNFile << endl;	
	}

	void evaluateMapping(PredicateMap& predsMapping, PredicateMap& bestMapping) {
		if (debug)
			cout << "Generate mapping file..." << endl;
		//output current mapping to file
		ofstream out;
		out.open(tmpPredMappingFile.c_str());	
		logOutput << endl << "Mapping " << mappingIndex++ << ":" << endl;
		PredicateMap::iterator it;
		for (it=predsMapping.begin(); it!= predsMapping.end(); it++) {
			const PredicateTemplate* currentSrcPred = (*it).first;
			out << currentSrcPred->getName() << " -> ";
			logOutput << currentSrcPred->getName() << " -> ";
			const PredicateTemplate* currentTargetPred = (*it).second;
			out << currentTargetPred->getName() << " ";
			logOutput << currentTargetPred->getName() << " ";
			out << endl;
			logOutput << endl;
		}
		out.close();
			
		if (debug) {
			cout << "Generate target MLN..." << endl;
			logOutput << "Generate target MLN..." << endl;
		}
		// generate target MLN based on the current predicate mapping
		sourceMLNToTargetMLN(tmpSrcMLNFile, predsMapping, tmpTargetMLNFile);
		
		// compute WPLL for the generated target MLN
		if (debug)
			cout << "Compute WPLL for the new generated MLN..." << endl;
		std::string inMLNFiles = tmpTargetPredicateFile;
		inMLNFiles.append(",").append(tmpTargetMLNFile);
		double sum = 0;
		int num = 0;
		Array<string> targetDbFilesArr;
		extractFileNames(targetDbFiles, targetDbFilesArr);
		
		if (debug) {
	  		cout << "Target database files: " << endl;
			for (int i = 0; i < targetDbFilesArr.size(); i++) {
				cout << targetDbFilesArr.item(i) << endl;
			}
	  	}
	
		for (int i = 0; i < targetDbFilesArr.size(); i++) {
			double wpll = computeWPLL(inMLNFiles.c_str(), targetDbFilesArr.item(i).c_str(), NULL, false);
			if (debug) cout << "WPLL: " << wpll << endl;
			if (wpll != 1) {
				sum += wpll;
				num++;
				logOutput << targetDbFilesArr.item(i) << ": " << wpll << endl;
			}
		}
		
		double score = -1e+10;
		if (num > 0) {
			score = sum / num;
			logOutput << "Average WPLL before revising: " << score << endl;
			//logOutput << endl;
		}
		
		if ((!doRevision) ||
		    (targetDbFilesArr.size() == 1)) 
		    {
				if (score > bestWPLLBeforeRevising) {
					bestWPLLBeforeRevising = score;
					bestWPLLAfterRevising = score;
					bestMappingIndex = mappingIndex - 1;
					bestMapping = predsMapping;
				}
		}
		else { // doRevision && targetDbFilesArr.size() > 1
			double lowerBound = (1 + confidentRatio)*bestWPLLBeforeRevising;
			if (score > lowerBound) {
				logOutput << "Do revising ... " << endl;
				
				double WPLLBeforeRevising = score;
				sum = 0;
				num = 0;
				const char* testDbFile = targetDbFilesArr.item(targetDbFilesArr.size()-1).c_str();
				std::string outRevisedMLNFile(tmpTargetMLNFile);
				outRevisedMLNFile.append("revision");
				for (int i = 0; i < targetDbFilesArr.size()-1; i++) {
					// do revising
					computeWPLL(inMLNFiles.c_str(), targetDbFilesArr.item(i).c_str(), outRevisedMLNFile.c_str(), true);
					
					// testing
					double wpll = computeWPLL(outRevisedMLNFile.c_str(), testDbFile, NULL, false);
					if (debug) cout << "WPLL after revising: " << wpll << endl;
					if (wpll != 1) {
						sum += wpll;
						num++;
						logOutput << targetDbFilesArr.item(i) << ": " << wpll << endl;
					}
				}
		
				double score = -1e+10;
				if (num > 0) {
					score = sum / num;
					logOutput << "Average WPLL after revising: " << score << endl;
					//logOutput << endl;
				}
						
				if (score > bestWPLLAfterRevising) {
					bestWPLLBeforeRevising = WPLLBeforeRevising;
					bestWPLLAfterRevising = score;
					bestMappingIndex = mappingIndex - 1;
					bestMapping = predsMapping;
				}
			}
		}
		
		
		// remove temporary files
		std::string command = "rm " + tmpPredMappingFile;
		system(command.c_str());
		command = "rm " + tmpTargetMLNFile; 
		system(command.c_str());
	
	}

	void globalPredicateMapping(const PredicateTemplateSet& sourcePreds, const PredicateTemplateSet& targetPreds, 
							int index, PredicateMap& predsMapping, StringMap typeConstraints[], PredicateMap& bestMapping) {
		if (index == sourcePreds.size()) {
			evaluateMapping(predsMapping, bestMapping);
		}
		else {
			if (index == 0) {
				predsMapping.clear();
			}
			const PredicateTemplate* currentSrcPred = sourcePreds[index];
			for (int i = 0; i < targetPreds.size(); i++) {
				const PredicateTemplate* currentTargetPred = targetPreds[i];
				if (isCompatible(currentSrcPred, currentTargetPred, typeConstraints)) {
					if (debug)
						logOutput << "compatilbe" << endl << endl;
					predsMapping[currentSrcPred] = currentTargetPred;
					//	add type constraints
					bool typeConstraintUpdates[currentSrcPred->getNumTerms()];
					for (int j = 0; j < currentSrcPred->getNumTerms(); j++) {
						const char* srcType = currentSrcPred->getTermTypeAsStr(j);
						const char* targetType = currentTargetPred->getTermTypeAsStr(j);
						const char* targetTypeConstraint = typeConstraints[0][srcType];
						if (targetTypeConstraint == NULL) { 
							typeConstraints[0][srcType] = targetType;
							typeConstraints[1][targetType] = srcType;
							typeConstraintUpdates[j] = true;
						}
						else
							typeConstraintUpdates[j] = false;
					}
					if (debug) {
						logOutput << "Adding type constraints..." << endl;
						logOutput << "Source to Target type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[0]);
						logOutput << "Target to Source type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[1]);
						logOutput << endl;
					}
					PredicateTemplateSet tmpTargetPreds(targetPreds);
					tmpTargetPreds.removeItem(i);
					globalPredicateMapping(sourcePreds,tmpTargetPreds,(index+1),predsMapping, typeConstraints, bestMapping);
					// remove type constraints
					for (int k = 0; k < currentSrcPred->getNumTerms(); k++) {
						if (typeConstraintUpdates[k]) {
							const char* srcType = currentSrcPred->getTermTypeAsStr(k); 
							const char* targetType = currentTargetPred->getTermTypeAsStr(k);						
							typeConstraints[0][srcType] = NULL;
							typeConstraints[1][targetType] = NULL;
						}
					}					
					if (debug) {
						logOutput << "Removing type constraints..." << endl;
						logOutput << "Source to Target type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[0]);
						logOutput << "Target to Source type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[1]);
						logOutput << endl;
					}
					
				}
				else {
					if (debug)
						logOutput << "not compatilbe" << endl << endl;
				}
			}
			PredicateTemplate* emptyPred = new PredicateTemplate();
			emptyPred->setName("");
			predsMapping[currentSrcPred] = emptyPred;
			globalPredicateMapping(sourcePreds,targetPreds,(index+1),predsMapping, typeConstraints, bestMapping);
		}
	}

	void findBestPredicateMappingForClause(const PredicateTemplateSet& sourcePreds, const PredicateTemplateSet& targetPreds, 
							int index, PredicateMap& predsMapping, StringMap typeConstraints[], PredicateMap& bestMapping)
	{
		if (index == sourcePreds.size()) {
			evaluateMapping(predsMapping, bestMapping);
		}
		else {
			//if (index == 0) {
				//predsMapping.clear();
			//}
			const PredicateTemplate* currentSrcPred = sourcePreds[index];
			for (int i = 0; i < targetPreds.size(); i++) {
				const PredicateTemplate* currentTargetPred = targetPreds[i];
				if (isCompatible(currentSrcPred, currentTargetPred, typeConstraints)) {
					if (debug)
						logOutput << "compatilbe" << endl << endl;
					predsMapping[currentSrcPred] = currentTargetPred;
					//update type constraints
					bool typeConstraintUpdates[currentSrcPred->getNumTerms()];
					for (int j = 0; j < currentSrcPred->getNumTerms(); j++) {
						const char* srcType = currentSrcPred->getTermTypeAsStr(j);
						const char* targetType = currentTargetPred->getTermTypeAsStr(j);
						const char* targetTypeConstraint = typeConstraints[0][srcType];
						if (targetTypeConstraint == NULL) { 
							typeConstraints[0][srcType] = targetType;
							typeConstraints[1][targetType] = srcType;
							typeConstraintUpdates[j] = true;
						}
						else
							typeConstraintUpdates[j] = false;
					}
					if (debug) {
						logOutput << "Updating type constraints..." << endl;
						logOutput << "Source to Target type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[0]);
						logOutput << "Target to Source type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[1]);
						logOutput << endl;
					}
					PredicateTemplateSet tmpTargetPreds(targetPreds);
					tmpTargetPreds.removeItem(i);
					findBestPredicateMappingForClause(sourcePreds,tmpTargetPreds,(index+1),predsMapping, typeConstraints, bestMapping);
					// remove type constraints
					for (int k = 0; k < currentSrcPred->getNumTerms(); k++) {
						if (typeConstraintUpdates[k]) {
							const char* srcType = currentSrcPred->getTermTypeAsStr(k); 
							const char* targetType = currentTargetPred->getTermTypeAsStr(k);						
							typeConstraints[0][srcType] = NULL;
							typeConstraints[1][targetType] = NULL;
						}
					}					
					if (debug) {
						logOutput << "Removing type constraints..." << endl;
						logOutput << "Source to Target type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[0]);
						logOutput << "Target to Source type constraints: " << endl;
						printTypeConstraints(logOutput, typeConstraints[1]);
						logOutput << endl;
					}
				}
				else {
					if (debug)
						logOutput << "not compatilbe" << endl << endl;
				}
			}
		}
	} 

	void clauseMapping(const Clause* sourceClause,const Domain* sourceDomain, const PredicateTemplateSet& targetPreds, StringMap typeConstraints[])
	{
		// ouput a source MLN containing only this clause
		ofstream out;
	  	out.open(tmpSrcMLNFile.c_str());
	  	sourceClause->printWithWtAndStrVar(out, sourceDomain);
	  	out.close();
	  	
	  	// find all predicates in this current source clause
		PredicateTemplateSet predsInSrcClause;
		for (int i = 0; i < sourceClause->getNumPredicates(); i++) {
			Predicate* pred = sourceClause->getPredicate(i);
			const PredicateTemplate* predTemplate = pred->getTemplate();
			if (debug)
				cout << (*predTemplate) << endl;
			if (!(predTemplate->isEqualPredWithType()))
	  			predsInSrcClause.append(predTemplate);
		}
		if (debug)
			cout << "Number of predicates in source clause: " << predsInSrcClause.size() << endl;
			
		// find the best mapping for this clause
		PredicateMap predsMapping;
		PredicateMap bestMapping;
		mappingIndex = 1;
		bestWPLLBeforeRevising = -1E+10;
		bestWPLLAfterRevising = -1E+10;
		findBestPredicateMappingForClause(predsInSrcClause, targetPreds, 0, predsMapping, typeConstraints, bestMapping);
		logOutput <<endl << "Best mapping is mapping " << bestMappingIndex << " with average WPLL = " << bestWPLLAfterRevising << endl;
		printClauseMapping(logOutput, sourceClause, sourceDomain, bestMapping);
		printClauseMapping(bestPredMappingOutput, sourceClause, sourceDomain, bestMapping);
		
		std::string tmpFile1("tmp-");
		tmpFile1.append(bestPredMappingFile);
		out.open(tmpFile1.c_str());
		printPredMapping(out, bestMapping);
		out.close();
		
		std::string tmpFile2("tmp-");
		tmpFile2.append(outTargetMLNFile);
		tmpFile2.append("-local");
		
		std::string command = "sourceMLNToTargetMLN.pl " + tmpSrcMLNFile + " " + tmpFile1 + " " + tmpFile2;
		//system(command.c_str());
		sourceMLNToTargetMLN(tmpSrcMLNFile, bestMapping, tmpFile2);
		
		copyFromFile(tmpFile2, bestTargetMLNOutput);
		
		// remove temporary files
		command = "rm " + tmpFile1;
		system(command.c_str());
		command = "rm " + tmpFile2;
		system(command.c_str());
	}

	void localPredicateMapping(const MLN* sourceMLN, const Domain* sourceDomain, const PredicateTemplateSet& sourcePreds,
								MLN* targetMLN, Domain* targetDomain, const PredicateTemplateSet& targetPreds,
								StringMap typeConstraints[]) 
	{
		if (debug) 
			cout << "Number of clauses in source domain: " << sourceMLN->getNumClauses() << endl;
		for (int i = 0; i < sourceMLN->getNumClauses(); i++) {
			const Clause* srcClause = sourceMLN->getClause(i);
			if (debug) {
				cout << "Clause " << (i+1) << ": ";
				srcClause->printWithWtAndStrVar(cout, sourceDomain);
				cout << endl;
			}
			
			logOutput << "Clause " << (i+1) << ": ";
			srcClause->printWithWtAndStrVar(logOutput, sourceDomain);
			logOutput << endl;
			
			if (srcClause->getNumPredicates() <= 1) {
				logOutput << "Do not transfer unit clause";
				continue;
			}
				
			clauseMapping(srcClause, sourceDomain, targetPreds, typeConstraints);
		}			
	}

	int file_select(const struct direct *entry)
	{
		if ((strcmp(entry->d_name, ".") == 0) ||
			(strcmp(entry->d_name, "..") == 0))
			return (FALSE);
		else
			return (TRUE);
	}

  void printTime1(ofstream& out, const double& sec)
  {
    if (sec < 60)    out << sec << " secs";
    else 
    if (sec < 3600)  out << int(sec/60) << " mins, " << sec-int(sec/60)*60 
                         << " secs";
    else 
    if (sec < 86400) out <<  int(sec/3600) << " hrs, " 
                         << (sec-int(sec/3600)*3600)/60.0 << " mins";
    else             out << int(sec/86400) << " days, " 
                         << (sec-int(sec/86400)*86400)/3600.0 << " hrs";
  }
 
 void trim2(string& str)
	{
	  string::size_type pos = str.find_last_not_of(' ');
	  if(pos != string::npos) {
	    str.erase(pos + 1);
	    pos = str.find_first_not_of(' ');
	    if(pos != string::npos) str.erase(0, pos);
	  }
	  else str.erase(str.begin(), str.end());
	}

	void printPredicates(const PredicateTemplateSet& preds) {
		for (int i = 0; i < preds.size(); i++) {
			const PredicateTemplate* currentPred = preds[i];
			cout << "Predicate " << (i+1) << ":";
			cout << (*currentPred) << endl;
		}
	}

	void printTypeConstraints(ostream& out, StringMap& typesConstraint) {
	  	out << "Number of constraints: " << typesConstraint.size() << endl;
	  	StringMap::iterator it;
		for (it=typesConstraint.begin(); it!= typesConstraint.end(); it++) {
			const char* srcType = (*it).first;
			if (srcType)
				out << srcType << " -> ";
			const char* targetType  = (*it).second;
			if (targetType)
				out << targetType << endl;
			else
				out << endl;
		}
	} 

	void printPredMapping(ostream& out, PredicateMap& predsMapping)
	{
		PredicateMap::iterator it;
		for (it=predsMapping.begin(); it!= predsMapping.end(); it++) {
			const PredicateTemplate* currentSrcPred = (*it).first;
			out << currentSrcPred->getName() << " -> ";
			const PredicateTemplate* currentTargetPred = (*it).second;
			out << currentTargetPred->getName() << " ";
			out << endl;
		}
	}

	void printMLNWithoutPredDeclaration(ostream& out, MLN* mln, Domain* domain) {
		for (int i = 0; i < mln->getNumClauses(); i++) {
			const Clause * currentClause = mln->getClause(i);
			currentClause->printWithWtAndStrVar(out, domain);
			out << endl;
		}
	}

	void printClauseMapping(ostream& out, const Clause* sourceClause,const Domain* sourceDomain, PredicateMap& bestMapping)
	{
		out << "Clause: ";
		sourceClause->printWithWtAndStrVar(out, sourceDomain);
		out << endl;
		printPredMapping(out, bestMapping);
		out << endl;
	}

	void copyFromFile(std::string sourceFile, ofstream& out) {
	  std::string line;
	  ifstream myfile(sourceFile.c_str());
	  if (myfile.is_open())
	  {
	    while (! myfile.eof() )
	    {
	      getline (myfile,line);
	      out << line << endl;
	    }
	    myfile.close();
	  }
	  else cout << "Unable to open file " << sourceFile.c_str() << endl;
	}
 
 private:
 	char* sourceMLNFile;
 	char* targetPredicateFile;
	char* targetDbFiles;
	char* mappingConstraintFile;
	char* outTargetMLNFile;
	bool globalMapping;
	bool sampling;
	bool debug;
	bool doRevision;
	int mappingIndex; // = 1;
	double bestWPLLBeforeRevising; // = -1E+10;
	double bestWPLLAfterRevising;
	double confidentRatio;
	int bestMappingIndex; // = 0;
	ofstream logOutput;
	char* funcFiles;
	std::string tmpSrcMLNFile;
	std::string tmpPredMappingFile;
	std::string tmpTargetMLNFile;
	std::string tmpTargetPredicateFile;
	std::string bestPredMappingFile;
	std::string logFile;
	ofstream bestPredMappingOutput;
	ofstream bestTargetMLNOutput;
	PredicateTemplateSet sourcePredicates;
	PredicateTemplateSet targetPredicates;
};


#endif /*PREDICATESMAPPING_H_*/
